"""Conversion functions for when datasets are missing particular variables and that xclim doesn't already implement."""

from __future__ import annotations  # for xclim 0.37

import xarray as xr
from xclim.core.units import convert_units_to, declare_units


@declare_units(prsn="[precipitation]", prlp="[precipitation]")
def precipitation(prsn: xr.DataArray, prlp: xr.DataArray) -> xr.DataArray:
    """Precipitation of all phases.

    Compute the precipitation flux from all phases by adding solid and liquid precipitation.

    Parameters
    ----------
    prsn: xr.DataArray
      Solid precipitation flux.
    prlp: xr.DataArray
      Liquid precipitation flux.

    Returns
    -------
    xr.DataArray, [same as prsn]
      Surface precipitation flux (all phases)
    """
    prlp = convert_units_to(prlp, prsn, context="hydro")
    pr = prsn + prlp
    pr.attrs["units"] = prsn.attrs["units"]
    return pr


@declare_units(dtr="[temperature]", tasmax="[temperature]")
def tasmin_from_dtr(dtr: xr.DataArray, tasmax: xr.DataArray) -> xr.DataArray:
    """Tasmin computed from DTR and tasmax.

    Tasmin as dtr subtracted from tasmax.

    Parameters
    ----------
    dtr: xr.DataArray
      Daily temperature range
    tasmax: xr.DataArray
      Daily maximal temperature.

    Returns
    -------
    xr.DataArray, [same as tasmax]
      Daily minimum temperature
    """
    out_units = tasmax.units
    # To prevent strange behaviors that could arise from changing DTR units, we change the units of tasmax to match DTR
    tasmax = convert_units_to(tasmax, dtr)
    tasmin = tasmax - dtr
    tasmin.attrs["units"] = dtr.units
    tasmin = convert_units_to(tasmin, out_units)
    return tasmin


@declare_units(dtr="[temperature]", tasmin="[temperature]")
def tasmax_from_dtr(dtr: xr.DataArray, tasmin: xr.DataArray) -> xr.DataArray:
    """Tasmax computed from DTR and tasmin.

    Tasmax as dtr added to tasmin.

    Parameters
    ----------
    dtr: xr.DataArray
      Daily temperature range
    tasmin: xr.DataArray
      Daily minimal temperature.

    Returns
    -------
    xr.DataArray, [same as tasmin]
      Daily maximum temperature
    """
    out_units = tasmin.attrs["units"]
    # To prevent strange behaviors that could arise from changing DTR units, we change the units of tasmin to match DTR
    tasmin = convert_units_to(tasmin, dtr)
    tasmax = tasmin + dtr
    tasmax.attrs["units"] = dtr.attrs["units"]
    tasmax = convert_units_to(tasmax, out_units)
    return tasmax


@declare_units(tasmin="[temperature]", tasmax="[temperature]")
def dtr_from_minmax(tasmin: xr.DataArray, tasmax: xr.DataArray) -> xr.DataArray:
    """DTR computed from tasmin and tasmax.

    Dtr as tasmin subtracted from tasmax.

    Parameters
    ----------
    tasmin: xr.DataArray
      Daily minimal temperature.
    tasmax: xr.DataArray
      Daily maximal temperature.

    Returns
    -------
    xr.DataArray
      Daily temperature range
    """
    tasmin = convert_units_to(tasmin, tasmax)
    dtr = tasmax - tasmin
    dtr.attrs["units"] = tasmin.attrs["units"]
    dtr.attrs["units_metadata"] = "temperature: difference"
    return dtr



"""Catalog objects and related tools."""

import ast
import itertools
import json
import logging
import os
import re
import shutil as sh
from collections.abc import Mapping, Sequence
from copy import deepcopy
from functools import reduce
from operator import or_
from pathlib import Path
from typing import Any

import fsspec as fs
import intake_esm
import pandas as pd
import tlz
import xarray as xr
from intake_esm.cat import ESMCatalogModel

from .config import CONFIG, args_as_str, recursive_update
from .utils import (
    _xarray_defaults,
    date_parser,
    ensure_correct_time,
    ensure_new_xrfreq,
    standardize_periods,
)

logger = logging.getLogger(__name__)
# Monkey patch for attribute names in the output of to_dataset_dict
intake_esm.set_options(attrs_prefix="cat")


__all__ = [
    "COLUMNS",
    "ID_COLUMNS",
    "DataCatalog",
    "ProjectCatalog",
    "concat_data_catalogs",
    "generate_id",
    "subset_file_coverage",
    "unstack_id",
]


# As much as possible, these catalog columns and entries should align with:
# https://github.com/WCRP-CMIP/CMIP6_CVs and https://github.com/ES-DOC/pyessv-archive
# See docs/columns.rst for a description of each entry.
COLUMNS = [
    "id",
    "type",
    "processing_level",
    "bias_adjust_institution",
    "bias_adjust_project",
    "mip_era",
    "activity",
    "driving_model",
    "institution",
    "source",
    "experiment",
    "member",
    "xrfreq",
    "frequency",
    "variable",
    "domain",
    "date_start",
    "date_end",
    "version",
    "format",
    "path",
]
"""Official column names."""

ID_COLUMNS = [
    "bias_adjust_project",
    "mip_era",
    "activity",
    "driving_model",
    "institution",
    "source",
    "experiment",
    "member",
    "domain",
]
"""Default columns used to create a unique ID"""


esm_col_data = {
    "esmcat_version": "0.1.0",  # intake-esm JSON file structure version, as per: https://github.com/NCAR/esm-collection-spec
    "assets": {"column_name": "path", "format_column_name": "format"},
    "aggregation_control": {
        "variable_column_name": "variable",
        "groupby_attrs": ["id", "domain", "processing_level", "xrfreq"],
        "aggregations": [
            {
                "type": "join_existing",
                "attribute_name": "date_start",
                "options": {"dim": "time"},
            },
            {"type": "union", "attribute_name": "variable"},
        ],
    },
    "attributes": [],
}
"""Default ESM column data for the official catalogs."""


def _parse_list_of_strings(elem):
    """Parse an element of a csv in case it is a tuple of strings."""
    if elem.startswith("(") or elem.startswith("["):
        out = ast.literal_eval(elem)
        return tuple(out)
    return (elem,)


def _parse_dates(elem):
    """Parse an array of dates (strings) into a PeriodIndex of hourly frequency."""
    # Cast to normal datetime as this is much faster than to period for in-bounds dates
    # errors are coerced to NaT, we convert to a PeriodIndex and then to a (mutable) series
    time = pd.to_datetime(elem, errors="coerce").astype(pd.PeriodDtype("H")).to_series()
    nat = time.isnull()
    # Only where we have NaT (parser errors and empty fields), parse into a Period
    # This will raise DateParseError as expected if the string is not parsable.
    time[nat] = pd.PeriodIndex(elem[nat], freq="H")
    return pd.PeriodIndex(time)


csv_kwargs = {
    "dtype": {
        key: "category" if not key == "path" else "string[pyarrow]"
        for key in COLUMNS
        if key not in ["xrfreq", "variable", "date_start", "date_end"]
    },
    "converters": {
        "variable": _parse_list_of_strings,
        "xrfreq": ensure_new_xrfreq,
    },
}
"""Kwargs to pass to `pd.read_csv` when opening an official Ouranos catalog."""


class DataCatalog(intake_esm.esm_datastore):
    r"""
    A read-only intake_esm catalog adapted to xscen's syntax.

    This class expects the catalog to have the columns listed in :py:data:`xscen.catalog.COLUMNS`
    and it comes with default arguments for reading the CSV files (:py:data:`xscen.catalog.csv_kwargs`).
    For example, all string columns (except `path`) are cast to a categorical dtype and the
    datetime columns are parsed with a special function that allows dates outside the conventional
    `datetime64[ns]` bounds by storing the data using :py:class:`pandas.Period` objects.

    Parameters
    ----------
    \*args : str or os.PathLike or dict
        Path to a catalog JSON file. If a dict, it must have two keys: 'esmcat' and 'df'.
        'esmcat' must be a dict representation of the ESM catalog.
        'df' must be a Pandas DataFrame containing content that would otherwise be in the CSV file.
    check_valid : bool
        If True, will check that all files in the catalog exist on disk and remove those that don't.
    drop_duplicates : bool
        If True, will drop duplicates in the catalog based on the 'id' and 'path' columns.
    \**kwargs : dict
        Any other arguments are passed to intake_esm.esm_datastore.

    See Also
    --------
    intake_esm.core.esm_datastore
    """

    def __init__(
        self, *args, check_valid: bool = False, drop_duplicates: bool = False, **kwargs
    ):
        kwargs["read_csv_kwargs"] = recursive_update(
            csv_kwargs.copy(), kwargs.get("read_csv_kwargs", {})
        )
        args = args_as_str(args)

        super().__init__(*args, **kwargs)

        # Cast date columns into datetime (with ms reso, that's why we do it here and not in the `read_csv_kwargs`)
        # Pandas >=2 supports [ms] resolution, but can't parse strings with this resolution, so we need to go through numpy
        for datecol in ["date_start", "date_end"]:
            if datecol in self.df.columns and self.df[datecol].dtype == "O":
                # Missing values in object columns are np.nan, which numpy can't convert to datetime64 (what's up with that numpy???)
                self.df[datecol] = self.df[datecol].dropna().astype("datetime64[ms]")

        if check_valid:
            self.check_valid()
        if drop_duplicates:
            self.drop_duplicates()

    @classmethod
    def from_df(
        cls,
        data: pd.DataFrame | os.PathLike | Sequence[os.PathLike],
        esmdata: os.PathLike | dict | None = None,
        *,
        read_csv_kwargs: Mapping[str, Any] | None = None,
        name: str = "virtual",
        **intake_kwargs,
    ):
        """Create a DataCatalog from one or more csv files.

        Parameters
        ----------
        data: DataFrame or path or sequence of paths
          A DataFrame or one or more paths to csv files.
        esmdata: path or dict, optional
          The "ESM collection data" as a path to a json file or a dict.
          If None (default), xscen's default :py:data:`esm_col_data` is used.
        read_csv_kwargs : dict, optional
          Extra kwargs to pass to `pd.read_csv`, in addition to the ones in :py:data:`csv_kwargs`.
        name: str
          If `metadata` doesn't contain it, a name to give to the catalog.

        See Also
        --------
        pandas.read_csv
        """
        if isinstance(data, pd.DataFrame):
            df = data
        else:
            if isinstance(data, os.PathLike):
                data = [data]

            read_csv_kwargs = recursive_update(csv_kwargs.copy(), read_csv_kwargs or {})

            df = pd.concat(
                [pd.read_csv(pth, **read_csv_kwargs) for pth in data]
            ).reset_index(drop=True)

        if isinstance(esmdata, os.PathLike):
            with Path(esmdata).open(encoding="utf-8") as f:
                esmdata = json.load(f)
        elif esmdata is None:
            esmdata = deepcopy(esm_col_data)
        if "id" not in esmdata:
            esmdata["id"] = name

        # Create the intake catalog
        return cls({"esmcat": esmdata, "df": df}, **intake_kwargs)

    def __dir__(self) -> list[str]:  # noqa: D105
        rv = ["iter_unique", "drop_duplicates", "check_valid"]
        return super().__dir__() + rv

    def _unique(self, columns) -> dict:
        def _find_unique(series):
            values = series.dropna()
            if series.name in self.esmcat.columns_with_iterables:
                values = tlz.concat(values)
            return list(tlz.unique(values))

        data = self.df[columns]
        if data.empty:
            return {col: [] for col in self.df.columns}
        else:
            return data.apply(_find_unique, result_type="reduce").to_dict()

    def unique(self, columns: str | Sequence[str] | None = None):
        """Return a series of unique values in the catalog.

        Parameters
        ----------
        columns : str or sequence of str, optional
          The columns to get unique values from. If None, all columns are used.
        """
        if self.df.size == 0:
            raise ValueError("Catalog is empty.")
        if isinstance(columns, str):
            cols = [columns]
        elif columns is None:
            cols = list(self.df.columns)
        else:
            cols = list(columns)
        uni = pd.Series(self._unique(cols))
        if isinstance(columns, str):
            return uni[columns]
        return uni

    def iter_unique(self, *columns):
        """Iterate over sub-catalogs for each group of unique values for all specified columns.

        This is a generator that yields a tuple of the unique values of the current
        group, in the same order as the arguments, and the sub-catalog.
        """
        for values in itertools.product(*self.unique(columns)):
            sim = self.search(**dict(zip(columns, values)))
            if sim:  # So we never yield empty catalogs
                yield values, sim

    def search(self, **columns):
        """Modification of .search() to add the 'periods' keyword."""
        periods = columns.pop("periods", False)
        if len(columns) > 0:
            cat = super().search(**columns)
        else:
            cat = self.__class__({"esmcat": self.esmcat.dict(), "df": self.esmcat._df})
        if periods is not False:
            cat.esmcat._df = subset_file_coverage(
                cat.esmcat._df, periods=periods, coverage=0, duplicates_ok=True
            )
        return cat

    def drop_duplicates(self, columns: list[str] | None = None):
        """Drop duplicates in the catalog based on a subset of columns.

        Parameters
        ----------
        columns: list of str, optional
            The columns used to identify duplicates. If None, 'id' and 'path' are used.
        """
        # In case variables are being added in an existing Zarr, append them
        if columns is None:
            columns = ["id", "path"]

        if len(self.df) > 0:
            # By default, duplicated will mark the later entries as True
            duplicated = self.df.duplicated(subset="path")
            df_dupes = self.df[duplicated]
            for _, d in df_dupes.iterrows():
                if Path(d.path).suffix == ".zarr":
                    append_v = list()
                    [
                        append_v.extend(v)
                        for v in self.df[self.df["path"] == d["path"]]["variable"]
                    ]
                    # Since setting multiple entries to tuples is a pain, update only the duplicated and re-add it to df
                    # Other entries will be dropped by drop_duplicates
                    d["variable"] = tuple(set(append_v))
                    self.esmcat._df = pd.concat(
                        [self.esmcat._df, pd.DataFrame(d).transpose()]
                    )

        # Drop duplicates
        self.esmcat.df.drop_duplicates(
            subset=columns, keep="last", ignore_index=True, inplace=True
        )

    def check_valid(self):
        """Verify that all files in the catalog exist on disk and remove those that don't.

        If a file is a Zarr, it will also check that all variables are present and remove those that aren't.
        """
        len_df = len(self.df)  # This line is required to avoid a D202 pydocstyle error

        # In case files were deleted manually, double-check that files do exist
        def check_existing(row):
            path = Path(row.path)
            exists = (path.is_dir() and path.suffix == ".zarr") or (path.is_file())
            if not exists:
                msg = f"File {path} was not found on disk, removing from catalog."
                logger.info(msg)
            return exists

        # In case variables were deleted manually in a Zarr, double-check that they still exist
        def check_variables(row):
            path = Path(row.path)
            if path.suffix == ".zarr":
                variables = [p.parts[-1] for p in path.iterdir()]
                exists = tuple(
                    set(
                        [row.variable]
                        if isinstance(row.variable, str)
                        else row.variable
                    ).intersection(variables)
                )
            else:
                exists = row.variable
            return exists

        if len_df > 0:
            self.esmcat._df = self.df[
                self.df.apply(check_existing, axis=1)
            ].reset_index(drop=True)
            if len_df > 0:
                self.esmcat._df["variable"] = self.df.apply(check_variables, axis=1)

    def exists_in_cat(self, **columns) -> bool:
        """
        Check if there is an entry in the catalogue corresponding to the arguments given.

        Parameters
        ----------
        columns: Arguments that will be given to `catalog.search`

        Returns
        -------
        bool
            True if there is an entry in the catalogue corresponding to the arguments given.
        """
        exists = bool(len(self.search(**columns)))
        if exists:
            msg = f"An entry exists for: {columns}"
            logger.info(msg)
        return exists

    def to_dataset(
        self,
        concat_on: list[str] | str | None = None,
        create_ensemble_on: list[str] | str | None = None,
        ensemble_name: list[str] | None = None,
        calendar: str | None = "standard",
        **kwargs,
    ) -> xr.Dataset:
        """
        Open the catalog's entries into a single dataset.

        Same as :py:meth:`~intake_esm.core.esm_datastore.to_dask`, but with additional control over the aggregations.
        The dataset definition logic is left untouched by this method (by default: ["id", "domain", "processing_level", "xrfreq"]),
        except that newly aggregated columns are removed from the "id".
        This will override any "custom" id, ones not unstackable with :py:func:`~xscen.catalog.unstack_id`.

        Ensemble preprocessing logic is taken from :py:func:`xclim.ensembles.create_ensemble`.
        When `create_ensemble_on` is given, the function ensures all entries have the correct time coordinate according to `xrfreq`.

        Parameters
        ----------
        concat_on : list of str or str, optional
          A list of catalog columns over which to concat the datasets (in addition to 'time').
          Each will become a new dimension with the column values as coordinates.
          Xarray concatenation rules apply and can be acted upon through `xarray_combine_by_coords_kwargs`.
        create_ensemble_on : list of str or str, optional
          The given column values will be merged into a new id-like "realization" column, which will be concatenated over.
          The given columns are removed from the dataset id, to remove them from the groupby_attrs logic.
          Xarray concatenation rules apply and can be acted upon through `xarray_combine_by_coords_kwargs`.
        ensemble_name : list of strings, optional
          If `create_ensemble_on` is given, this can be a subset of those column names to use when constructing the realization coordinate.
          If None, this will be the same as `create_ensemble_on`.
          The resulting coordinate must be unique.
        calendar : str, optional
          If `create_ensemble_on` is given but not `preprocess`, all datasets are converted to this calendar before concatenation.
          Ignored otherwise (default). If None, no conversion is done. `align_on` is always "date".
          If `preprocess` is given, it must do the needed calendar handling.
        kwargs:
          Any other arguments are passed to :py:meth:`~intake_esm.core.esm_datastore.to_dataset_dict`.
          The `preprocess` argument must convert calendars as needed if `create_ensemble_on` is given.

        Returns
        -------
        :py:class:`~xarray.Dataset`

        See Also
        --------
        intake_esm.core.esm_datastore.to_dataset_dict
        intake_esm.core.esm_datastore.to_dask
        xclim.ensembles.create_ensemble
        """
        cat = deepcopy(self)
        # Put back what was removed by the copy...
        cat._requested_variables = self._requested_variables

        if isinstance(concat_on, str):
            concat_on = [concat_on]
        if isinstance(create_ensemble_on, str):
            create_ensemble_on = [create_ensemble_on]
        if ensemble_name is None:
            ensemble_name = create_ensemble_on
        elif not set(ensemble_name).issubset(create_ensemble_on):
            raise ValueError(
                "`ensemble_name` must be a subset of `create_ensemble_on`."
            )
        rm_from_id = (concat_on or []) + (create_ensemble_on or []) + ["realization"]

        aggs = {
            agg.attribute_name for agg in cat.esmcat.aggregation_control.aggregations
        }
        if not set(cat.esmcat.aggregation_control.groupby_attrs).isdisjoint(rm_from_id):
            raise ValueError(
                "Can't add aggregations for columns in the catalog's groupby_attrs "
                f"({cat.esmcat.aggregation_control.groupby_attrs})"
            )
        if not aggs.isdisjoint(rm_from_id):
            raise ValueError(
                f"Can't add aggregations for columns were an aggregation already exists ({aggs})"
            )

        if concat_on:
            cat.esmcat.aggregation_control.aggregations.extend(
                [
                    intake_esm.cat.Aggregation(
                        type=intake_esm.cat.AggregationType.join_new, attribute_name=col
                    )
                    for col in concat_on
                ]
            )

        if create_ensemble_on:
            cat.df["realization"] = generate_id(cat.df, ensemble_name)
            cat.esmcat.aggregation_control.aggregations.append(
                intake_esm.cat.Aggregation(
                    type=intake_esm.cat.AggregationType.join_new,
                    attribute_name="realization",
                )
            )
            xrfreq = cat.df["xrfreq"].unique()[0]

            if kwargs.get("preprocess") is None:

                def preprocess(ds):
                    ds = ensure_correct_time(ds, xrfreq)
                    if calendar is not None:
                        ds = ds.convert_calendar(
                            calendar,
                            use_cftime=(calendar != "default"),
                            align_on="date",
                        )
                    return ds

                kwargs["preprocess"] = preprocess

        if len(rm_from_id) > 1:
            # Guess what the ID was and rebuild a new one, omitting the columns part of the aggregation
            unstacked = unstack_id(cat)
            cat.esmcat.df["id"] = cat.df.apply(
                lambda row: _build_id(
                    row, [col for col in unstacked[row["id"]] if col not in rm_from_id]
                ),
                axis=1,
            )

        if (N := len(cat.keys())) != 1:
            raise ValueError(
                f"Expected exactly one dataset, received {N} instead : {cat.keys()}"
            )

        kwargs = _xarray_defaults(**kwargs)

        ds = cat.to_dask(**kwargs)
        return ds

    def copy_files(
        self,
        dest: str | os.PathLike,
        flat: bool = True,
        unzip: bool = False,
        zipzarr: bool = False,
        inplace: bool = False,
    ):
        """Copy each file of the catalog to another location, unzipping datasets along the way if requested.

        Parameters
        ----------
        cat: DataCatalog or ProjectCatalog
            A catalog to copy.
        dest: str, path
            The root directory of the destination.
        flat: bool
            If True (default), all dataset files are copied in the same directory.
            Renaming with an integer suffix ("{name}_01.{ext}") is done in case of duplicate file names.
            If False, :py:func:`xscen.catutils.build_path` (with default arguments) is used to generated the new path below the destination.
            Nothing is done in case of duplicates in that case.
        unzip: bool
            If True, any datasets with a `.zip` suffix are unzipped during the copy (or rather instead of a copy).
        zipzarr: bool
            If True, any datasets with a `.zarr` suffix are zipped during the copy (or rather instead of a copy).
        inplace : bool
            If True, the catalog is updated in place. If False (default), a copy is returned.

        Returns
        -------
        If inplace is False, this returns a catalog similar to self except with updated filenames. Some special attributes are not preserved,
        such as those added by :py:func:`xscen.extract.search_data_catalogs`. In this case, use `inplace=True`.
        """
        # Local imports to avoid circular imports
        from .catutils import build_path
        from .io import unzip_directory, zip_directory

        dest = Path(dest)
        data = self.esmcat._df.copy()
        if flat:
            new_paths = []
            for path in map(Path, data.path.values):
                if unzip and path.suffix == ".zip":
                    new = dest / path.with_suffix("").name
                elif zipzarr and path.suffix == ".zarr":
                    new = dest / path.with_suffix(".zarr.zip").name
                else:
                    new = dest / path.name
                if new in new_paths:
                    suffixes = "".join(new.suffixes)
                    name = new.name.removesuffix(suffixes)
                    i = 1
                    while new in new_paths:
                        new = dest / (name + f"_{i:02d}" + suffixes)
                        i += 1
                new_paths.append(new)
            data["new_path"] = new_paths
        else:
            data = build_path(data, root=dest).drop(columns=["new_path_type"])

        msg = f"Will copy {len(data)} files."
        logger.debug(msg)
        for i, row in data.iterrows():
            old = Path(row.path)
            new = Path(row.new_path)
            if unzip and old.suffix == ".zip":
                msg = f"Unzipping {old} to {new}."
                logger.info(msg)
                unzip_directory(old, new)
            elif zipzarr and old.suffix == ".zarr":
                msg = f"Zipping {old} to {new}."
                logger.info(msg)
                zip_directory(old, new)
            elif old.is_dir():
                msg = f"Copying directory tree {old} to {new}."
                logger.info(msg)
                sh.copytree(old, new)
            else:
                msg = f"Copying file {old} to {new}."
                logger.info(msg)
                sh.copy(old, new)
        if inplace:
            self.esmcat._df["path"] = data["new_path"]
            return
        data["path"] = data["new_path"]
        data = data.drop(columns=["new_path"])
        return self.__class__({"esmcat": self.esmcat.model_dump(), "df": data})


class ProjectCatalog(DataCatalog):
    """A DataCatalog with additional 'write' functionalities that can update and upload itself.

    See Also
    --------
    intake_esm.core.esm_datastore
    """

    @classmethod
    def create(
        cls,
        filename: os.PathLike | str,
        *,
        project: dict | None = None,
        overwrite: bool = False,
    ):
        r"""Create a new project catalog from some project metadata.

        Creates the json from default :py:data:`esm_col_data` and an empty csv file.

        Parameters
        ----------
        filename : os.PathLike or str
            A path to the json file (with or without suffix).
        project : dict, optional
            Metadata to create the catalog. If None, `CONFIG['project']` will be used.
            Valid fields are:

            - title : Name of the project, given as the catalog's "title".
            - id : slug-like version of the name, given as the catalog's id (should be url-proof)
              Defaults to a modified name.
            - version : Version of the project (and thus the catalog), string like "x.y.z".
            - description : Detailed description of the project, given to the catalog's "description".
            - Any other entry defined in :py:data:`esm_col_data`.

            At least one of `id` and `title` must be given, the rest is optional.
        overwrite : bool
            If True, will overwrite any existing JSON and CSV file.

        Returns
        -------
        ProjectCatalog
            An empty intake_esm catalog.
        """
        path = Path(filename)
        meta_path = path.with_suffix(".json")
        data_path = path.with_suffix(".csv")

        if (meta_path.is_file() or data_path.is_file()) and not overwrite:
            raise FileExistsError(
                "Catalog file already exist (at least one of {meta_path} or {data_path})."
            )

        meta_path.parent.mkdir(parents=True, exist_ok=True)

        project = project or CONFIG.get("project") or {}

        if "id" not in project and "title" not in project:
            raise ValueError(
                'At least one of "id" or "title" must be given in the metadata.'
            )

        project["catalog_file"] = str(data_path)
        if "id" not in project:
            project["id"] = project.get("title", "").replace(" ", "")

        esmdata = recursive_update(esm_col_data.copy(), project)

        df = pd.DataFrame(columns=COLUMNS)

        cat = cls(
            {"esmcat": esmdata, "df": df}
        )  # TODO: Currently, this drops "version" because it is not a recognized attribute
        cat.serialize(
            path.stem,
            directory=path.parent,
            catalog_type="file",
            to_csv_kwargs={"compression": None},
        )

        # Change catalog_file to a relative path
        with Path(meta_path).open(encoding="utf-8") as f:
            meta = json.load(f)
            meta["catalog_file"] = data_path.name
        with Path(meta_path).open("w", encoding="utf-8") as f:
            json.dump(meta, f, indent=2)

        return cls(str(meta_path))

    def __init__(
        self,
        df: str | dict,
        *args,
        create: bool = False,
        overwrite: bool = False,
        project: dict | None = None,
        check_valid: bool = True,
        drop_duplicates: bool = True,
        **kwargs,
    ):
        r"""
        Open or create a project catalog.

        Parameters
        ----------
        df : str or dict
            If str, this must be a path or URL to a catalog JSON file.
            If dict, this must be a dict representation of an ESM catalog.  See the notes below.
        create : bool
            If True, and if 'df' is a string, this will create an empty ProjectCatalog if none already exists.
        overwrite : bool
            If this and 'create' are True, this will overwrite any existing JSON and CSV file with an empty catalog.
        project : dict, optional
            Metadata to create the catalog, if required.
        check_valid : bool
            If True (default), will check that all files in the catalog exist on disk and remove those that don't.
        drop_duplicates : bool
            If True (default), will drop duplicates in the catalog based on the 'id' and 'path' columns.
        \**kwargs : dict
            Any other arguments are passed to xscen.catalog.DataCatalog.


        Notes
        -----
        New ProjectCatalog must first be created empty, using 'df' as the path to the future JSON file, then populated using .update().
        The dictionary in 'df' must have two keys: ‘esmcat’ and ‘df’.
        The ‘esmcat’ key must be a dict representation of the ESM catalog. This should follow the template used by xscen.catalog.esm_col_data.
        The ‘df’ key must be a Pandas DataFrame containing content that would otherwise be in the CSV file.
        """
        if create:
            if isinstance(df, str | Path) and (not Path(df).is_file() or overwrite):
                self.create(df, project=project, overwrite=overwrite)
        super().__init__(
            df,
            *args,
            check_valid=check_valid,
            drop_duplicates=drop_duplicates,
            **kwargs,
        )
        self.check_valid_flag = check_valid
        self.drop_duplicates_flag = drop_duplicates
        self.meta_file = df if not isinstance(df, dict) else None

    # TODO: Implement a way to easily destroy part of the catalog to "reset" some steps
    def update(
        self,
        df: None | (
            DataCatalog
            | intake_esm.esm_datastore
            | pd.DataFrame
            | pd.Series
            | Sequence[pd.Series]
        ) = None,
    ):
        """Update the catalog with new data and writes the new data to the csv file.

        Once the internal dataframe is updated with `df`, the csv on disk is parsed,
        updated with the internal dataframe, duplicates are dropped and everything is
        written back to the csv. This means that nothing is _removed_* from the csv when
        calling this method, and it is safe to use even with a subset of the catalog.

        Warnings
        --------
        If a file was deleted between the parsing of the catalog and this call,
        it will be removed from the csv if `check_valid` is called.

        Parameters
        ----------
        df : DataCatalog | intake_esm.esm_datastore | pd.DataFrame | pd.Series  | Sequence[pd.Series], optional
            Data to be added to the catalog. If None, nothing is added, but the catalog is still updated.
        """
        # Append the new DataFrame or Series
        if isinstance(df, DataCatalog) or isinstance(df, intake_esm.esm_datastore):
            self.esmcat._df = pd.concat([self.df, df.df])
        elif df is not None:
            if isinstance(df, pd.Series):
                df = pd.DataFrame(df).transpose()
            self.esmcat._df = pd.concat([self.df, df])
        if self.check_valid_flag:
            self.check_valid()
        if self.drop_duplicates_flag:
            self.drop_duplicates()

        # make sure year really has 4 digits
        if "date_start" in self.df:
            df_fix_date = self.df.copy()
            df_fix_date["date_start"] = pd.Series(
                [
                    (
                        x
                        if isinstance(x, str)
                        else "" if pd.isnull(x) else x.strftime("%4Y-%m-%d %H:00")
                    )
                    for x in self.df.date_start
                ]
            )

            df_fix_date["date_end"] = pd.Series(
                [
                    (
                        x
                        if isinstance(x, str)
                        else "" if pd.isnull(x) else x.strftime("%4Y-%m-%d %H:00")
                    )
                    for x in self.df.date_end
                ]
            )

            df_str = df_fix_date
        else:
            df_str = self.df

        if self.meta_file is not None:
            with fs.open(self.esmcat.catalog_file, "wb") as csv_outfile:
                df_str.to_csv(csv_outfile, index=False, compression=None)
        else:
            read_csv_kwargs = deepcopy(self.read_csv_kwargs)
            del read_csv_kwargs["parse_dates"]
            del read_csv_kwargs["date_parser"]
            # Update the catalog file saved on disk
            disk_cat = DataCatalog(
                {
                    "esmcat": self.esmcat.dict(),
                    "df": pd.read_csv(self.esmcat.catalog_file, **read_csv_kwargs),
                }
            )
            disk_cat.esmcat._df = pd.concat([disk_cat.df, df_str])
            if self.check_valid_flag:
                disk_cat.check_valid()
            if self.drop_duplicates_flag:
                disk_cat.drop_duplicates()
            with fs.open(disk_cat.esmcat.catalog_file, "wb") as csv_outfile:
                disk_cat.df.to_csv(csv_outfile, index=False, compression=None)

    def update_from_ds(
        self,
        ds: xr.Dataset,
        path: os.PathLike | str,
        info_dict: dict | None = None,
        **info_kwargs,
    ):
        """Update the catalog with new data and writes the new data to the csv file.

        We get the new data from the attributes of `ds`, the dictionary `info_dict` and `path`.

        Once the internal dataframe is updated with the new data, the csv on disk is parsed,
        updated with the internal dataframe, duplicates are dropped and everything is
        written back to the csv. This means that nothing is _removed_* from the csv when
        calling this method, and it is safe to use even with a subset of the catalog.

        Warnings
        --------
        If a file was deleted between the parsing of the catalog and this call,
        it will be removed from the csv if `check_valid` is called.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset that we want to add to the catalog.
            The columns of the catalog will be filled from the global attributes starting with 'cat:' of the dataset.
        info_dict : dict, optional
            Extra information to fill in the catalog.
        path : os.PathLike or str
            Path to the file that contains the dataset. This will be added to the 'path' column of the catalog.
        """
        d = {}

        for col in self.df.columns:
            if f"cat:{col}" in ds.attrs:
                d[col] = ds.attrs[f"cat:{col}"]
        if info_dict:
            d.update(info_dict)
        if info_kwargs:
            d.update(info_kwargs)

        if "time" in ds.dims:
            d["date_start"] = str(
                ds.isel(time=0).time.dt.strftime("%4Y-%m-%d %H:%M:%S").values
            )
            d["date_end"] = str(
                ds.isel(time=-1).time.dt.strftime("%4Y-%m-%d %H:%M:%S").values
            )

        d["path"] = str(Path(path))

        # variable should be based on the Dataset
        d["variable"] = tuple(v for v in ds.data_vars if len(ds[v].dims) > 0)

        if "format" not in d:
            d["format"] = Path(d["path"]).suffix.split(".")[1]
            msg = f"File format not specified. Adding it as '{d['format']}' based on file name."
            logger.info(msg)

        self.update(pd.Series(d))

    def refresh(self):
        """Reread the catalog CSV saved on disk."""
        if self.meta_file is None:
            raise ValueError(
                "Only full catalogs can be refreshed, but this instance is only a subset."
            )
        self.esmcat = ESMCatalogModel.load(
            self.meta_file, read_csv_kwargs=self.read_csv_kwargs
        )
        initlen = len(self.esmcat.df)
        if self.check_valid_flag:
            self.check_valid()
        if self.drop_duplicates_flag:
            self.drop_duplicates()
        if len(self.df) != initlen:
            self.update()

    def __repr__(self) -> str:  # noqa: D105
        return (
            f'<{self.esmcat.id or ""} project catalog with {len(self)} dataset(s) from '
            f'{len(self.df)} asset(s) ({"subset" if self.meta_file is None else "full"})>'
        )


def concat_data_catalogs(*dcs):
    """Concatenate a multiple DataCatalogs.

    Output catalog is the union of all rows and all derived variables, with the "esmcat"
    of the first DataCatalog. Duplicate rows are dropped and the index is reset.
    """
    registry = {}
    catalogs = []
    requested_variables = []
    requested_variables_true = []
    dependent_variables = []
    requested_variable_freqs = []
    for dc in dcs:
        registry.update(dc.derivedcat._registry)
        catalogs.append(dc.df)
        requested_variables.extend(dc._requested_variables)
        requested_variables_true.extend(getattr(dc, "_requested_variables_true", []))
        dependent_variables.extend(getattr(dc, "_dependent_variables", []))
        requested_variable_freqs.extend(getattr(dc, "_requested_variable_freqs", []))
    df = pd.concat(catalogs, axis=0).drop_duplicates(ignore_index=True)
    dvr = intake_esm.DerivedVariableRegistry()
    dvr._registry.update(registry)
    newcat = DataCatalog({"esmcat": dcs[0].esmcat.model_dump(), "df": df}, registry=dvr)
    newcat._requested_variables = requested_variables
    if requested_variables_true:
        newcat._requested_variables_true = requested_variables_true
    if dependent_variables:
        newcat._dependent_variables = dependent_variables
    if requested_variable_freqs:
        newcat._requested_variable_freqs = requested_variable_freqs
    return newcat


def _build_id(element: pd.Series, columns: list[str]):
    """Build an ID from a catalog's row and a list of columns."""
    return "_".join(map(str, filter(pd.notna, element[columns].values)))


def generate_id(
    df: pd.DataFrame | xr.Dataset, id_columns: list | None = None
) -> pd.Series:
    """Create an ID from column entries.

    Parameters
    ----------
    df: pd.DataFrame, xr.Dataset
      Data for which to create an ID.
    id_columns : list, optional
      List of column names on which to base the dataset definition. Empty columns will be skipped.
      If None (default), uses :py:data:`ID_COLUMNS`.

    Returns
    -------
    pd.Series
        A series of IDs, one per row of the input DataFrame.
    """
    if isinstance(df, xr.Dataset):
        df = pd.DataFrame.from_dict(
            {
                key[4:]: [value]
                for key, value in df.attrs.items()
                if key.startswith("cat:")
            }
        )

    id_columns = [x for x in (id_columns or ID_COLUMNS) if x in df.columns]

    return df.apply(_build_id, axis=1, args=(id_columns,))


def unstack_id(df: pd.DataFrame | ProjectCatalog | DataCatalog) -> dict:
    """Reverse-engineer an ID using catalog entries.

    Parameters
    ----------
    df : pd.DataFrame or ProjectCatalog or DataCatalog
        Either a Project/DataCatalog or a pandas DataFrame.

    Returns
    -------
    dict
        Dictionary with one entry per unique ID, which are themselves dictionaries of all the individual parts of the ID.
    """
    if isinstance(df, ProjectCatalog | DataCatalog):
        df = df.df

    out = {}
    for ids in pd.unique(df["id"]):
        subset = df[df["id"] == ids]

        # Only keep relevant columns
        subset = subset[
            [
                col
                for col in subset.columns
                if bool(re.search(f"((_)|(^)){subset[col].iloc[0]!s}((_)|($))", ids))
            ]
        ].drop("id", axis=1)

        # Make sure that all elements are the same, if there are multiple lines
        if not (subset.nunique() == 1).all():
            raise ValueError(
                "Not all elements of the columns are the same for a given ID!"
            )

        out[ids] = {attr: subset[attr].iloc[0] for attr in subset.columns}

    return out


def subset_file_coverage(
    df: pd.DataFrame,
    periods: list[str] | list[list[str]],
    *,
    coverage: float = 0.99,
    duplicates_ok: bool = False,
) -> pd.DataFrame:
    """Return a subset of files that overlap with the target periods.

    Parameters
    ----------
    df : pd.DataFrame
        List of files to be evaluated, with at least a date_start and date_end column,
        which are expected to be `datetime64` objects.
    periods : list of str or list of lists of str
        Either [start, end] or list of [start, end] for the periods to be evaluated.
        All periods must be covered, otherwise an empty subset is returned.
    coverage : float
        Percentage of hours that need to be covered in a given period for the dataset to be valid. Use 0 to ignore this checkup.
        The coverage calculation is only valid if there are no overlapping periods in `df` (ensure with `duplicates_ok=False`).
    duplicates_ok : bool
        If True, no checkup is done on possible duplicates.

    Returns
    -------
    pd.DataFrame
        Subset of files that overlap the targeted periods.
    """
    periods = standardize_periods(periods, out_dtype="datetime")

    # Create an Interval for each file
    intervals = pd.IntervalIndex.from_arrays(
        df["date_start"], df["date_end"], closed="both"
    )

    # Check for duplicated Intervals
    if duplicates_ok is False and intervals.is_overlapping:
        msg = f"{df['id'].iloc[0] + ': ' if 'id' in df.columns else ''}Time periods are overlapping."
        logging.warning(msg)
        return pd.DataFrame(columns=df.columns)

    # Create an array of True/False
    files_to_keep = []
    for period in periods:
        period_interval = pd.Interval(
            date_parser(period[0]),
            date_parser(period[1], end_of_period=True),
            closed="both",
        )
        files_in_range = intervals.overlaps(period_interval)

        if not files_in_range.any():
            msg = f"{df['id'].iloc[0] + ': ' if 'id' in df.columns else ''}Insufficient coverage (no files in range {period})."
            logging.warning(msg)
            return pd.DataFrame(columns=df.columns)

        # Very rough guess of the coverage relative to the requested period,
        # without having to open the files or checking day-by-day
        if coverage > 0:
            # Number of hours in the requested period
            period_length = period_interval.length
            # Sum of hours in all selected files, restricted by the requested period
            guessed_length = pd.IntervalIndex.from_arrays(
                intervals[files_in_range].map(
                    lambda x: max(x.left, period_interval.left)
                ),
                intervals[files_in_range].map(
                    lambda x: min(x.right, period_interval.right)
                ),
            ).length.sum()

            if guessed_length / period_length < coverage:
                msg = (
                    f"{df['id'].iloc[0] + ': ' if 'id' in df.columns else ''}Insufficient coverage "
                    f"(guessed at {guessed_length / period_length:.1%})."
                )
                logging.warning(msg)
                return pd.DataFrame(columns=df.columns)

        files_to_keep.append(files_in_range)

    return df[reduce(or_, files_to_keep)]



from pathlib import Path

import pytest

import xscen as xs


class TestPublish:
    @pytest.mark.requires_netcdf
    @pytest.mark.parametrize("fmt", ["md", "rst"])
    def test_normal(self, fmt):
        out = xs.testing.publish_release_notes(
            fmt,
            changes=Path(__file__).parent.parent.joinpath("CHANGELOG.rst"),
            latest=False,
        )
        if fmt == "md":
            assert out.startswith("# Changelog\n\n")
            assert "[PR/413](https://github.com/Ouranosinc/xscen/pull/413)" in out
        elif fmt == "rst":
            assert out.startswith("=========\nChangelog\n=========\n\n")
            assert "`PR/413 <https://github.com/Ouranosinc/xscen/pull/\\>`_" in out

    def test_error(self):
        with pytest.raises(FileNotFoundError):
            xs.testing.publish_release_notes("md", changes="foo")
        with pytest.raises(NotImplementedError):
            xs.testing.publish_release_notes(
                "foo", changes=Path(__file__).parent.parent.joinpath("CHANGELOG.rst")
            )

    @pytest.mark.requires_netcdf
    def test_file(self, tmpdir):
        xs.testing.publish_release_notes(
            "md",
            file=tmpdir / "foo.md",
            changes=Path(__file__).parent.parent.joinpath("CHANGELOG.rst"),
        )
        with Path(tmpdir).joinpath("foo.md").open(encoding="utf-8") as f:
            assert f.read().startswith("# Changelog\n\n")

    @pytest.mark.parametrize("latest", [True, False])
    @pytest.mark.requires_netcdf
    def test_latest(self, tmpdir, latest):
        out = xs.testing.publish_release_notes(
            "md",
            changes=Path(__file__).parent.parent.joinpath("CHANGELOG.rst"),
            latest=latest,
        )
        if latest:
            assert len(out.split("\n\n## v0.")) == 2
        else:
            assert len(out.split("\n\n## v0.")) > 2


def test_show_version(tmpdir):
    xs.testing.show_versions(file=tmpdir / "versions.txt")
    with Path(tmpdir).joinpath("versions.txt").open(encoding="utf-8") as f:
        out = f.read()
    assert "xscen" in out
    assert "xclim" in out
    assert "xesmf" in out
    assert "xarray" in out
    assert "numpy" in out
    assert "pandas" in out
    assert "dask" in out
    assert "cftime" in out
    assert "netCDF4" in out



"""Functions to regrid datasets."""

import datetime
import operator
import os
import random
import string
import warnings
from copy import deepcopy
from pathlib import Path

import cartopy.crs as ccrs
import cf_xarray as cfxr
import xarray as xr
from xclim.core.units import convert_units_to

try:
    import xesmf as xe
    from xesmf.frontend import Regridder
except ImportError:
    xe = None
    Regridder = "xesmf.Regridder"

from .config import parse_config
from .spatial import get_grid_mapping

__all__ = ["create_bounds_gridmapping", "create_mask", "regrid_dataset"]


@parse_config
def regrid_dataset(  # noqa: C901
    ds: xr.Dataset,
    ds_grid: xr.Dataset,
    *,
    weights_location: str | os.PathLike | None = None,
    regridder_kwargs: dict | None = None,
    intermediate_grids: dict | None = None,
    to_level: str = "regridded",
) -> xr.Dataset:
    """Regrid a dataset according to weights and a reference grid.

    Based on an intake_esm catalog, this function performs regridding on Zarr files.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to regrid. The Dataset needs to have lat/lon coordinates.
        Supports a 'mask' variable compatible with ESMF standards.
    ds_grid : xr.Dataset
        Destination grid. The Dataset needs to have lat/lon coordinates.
        Supports a 'mask' variable compatible with ESMF standards.
    weights_location : Union[str, os.PathLike], optional
        Path to the folder where weight file is saved. Leave as None to force re-computation of weights.
        Note that in order to reuse the weights, ds and ds_grid should both have the 'cat:id' and 'cat:domain' attributes.
    regridder_kwargs : dict, optional
        Arguments to send xe.Regridder(). If it contains `skipna` or `output_chunks`, those
        are passed to the regridder call directly.
    intermediate_grids : dict, optional
        This argument is used to do a regridding in many steps, regridding to regular
        grids before regridding to the final ds_grid.
        This is useful when there is a large jump in resolution between ds and ds grid.
        The format is a nested dictionary shown in Notes.
        If None, no intermediary grid is used, there is only a regrid from ds to ds_grid.
    to_level : str
        The processing level to assign to the output.
        Defaults to 'regridded'

    Returns
    -------
    xarray.Dataset
        Regridded dataset

    Notes
    -----
    intermediate_grids =
      {'name_of_inter_grid_1': {'cf_grid_2d': {arguments for util.cf_grid_2d },'regridder_kwargs':{arguments for xe.Regridder}},
        'name_of_inter_grid_2': dictionary_as_above}

    See Also
    --------
    xesmf.regridder, xesmf.util.cf_grid_2d
    """
    if xe is None:
        raise ImportError(
            "xscen's regridding functionality requires xESMF to work, please install that package."
        )
    ds = ds.copy()
    regridder_kwargs = regridder_kwargs or {}

    # We modify the dataset later, so we need to keep track of whether it had lon_bounds and lat_bounds to begin with
    has_lon_bounds = "lon_bounds" in ds
    has_lat_bounds = "lat_bounds" in ds

    # Generate unique IDs to name the weights file, but remove the member and experiment from the dataset ID
    if weights_location is not None:
        dsid = (
            ds.attrs.get("cat:id", _generate_random_string(15))
            .replace(ds.attrs.get("cat:member", ""), "")
            .replace(ds.attrs.get("cat:experiment", ""), "")
        )
        dsid = f"{dsid}_{ds.attrs.get('cat:domain', _generate_random_string(15))}"
        gridid = f"{ds_grid.attrs.get('cat:id', _generate_random_string(15))}_{ds_grid.attrs.get('cat:domain', _generate_random_string(15))}"

    ds_grids = []  # List of target grids
    reg_arguments = []  # List of accompanying arguments for xe.Regridder()
    if intermediate_grids:
        for name_inter, dict_inter in intermediate_grids.items():
            reg_arguments.append(dict_inter["regridder_kwargs"])
            ds_grids.append(xe.util.cf_grid_2d(**dict_inter["cf_grid_2d"]))

    ds_grids.append(ds_grid)  # Add the final ds_grid
    reg_arguments.append(regridder_kwargs)  # Add the final regridder_kwargs

    out = None

    # If the grid is the same, skip the call to xESMF
    if ds["lon"].equals(ds_grid["lon"]) & ds["lat"].equals(ds_grid["lat"]):
        out = ds
        if "mask" in out:
            out = out.where(out.mask == 1)
            out = out.drop_vars(["mask"])
        if "mask" in ds_grid:
            out = out.where(ds_grid.mask == 1)

    else:
        for i, (ds_grid, regridder_kwargs) in enumerate(zip(ds_grids, reg_arguments)):
            # If this is not the first iteration (out != None),
            # get the result from the last iteration (out) as input
            ds = out or ds
            kwargs = deepcopy(regridder_kwargs)

            # Prepare the weight file
            if weights_location is not None:
                Path(weights_location).mkdir(parents=True, exist_ok=True)
                weights_filename = Path(
                    weights_location,
                    f"{dsid}_{gridid}_regrid{i}"
                    f"{'_'.join(kwargs[k] for k in kwargs if isinstance(kwargs[k], str))}.nc",
                )

                # Reuse existing weight file if possible
                if Path(weights_filename).is_file() and not (
                    ("reuse_weights" in kwargs) and (kwargs["reuse_weights"] is False)
                ):
                    kwargs["weights"] = weights_filename
                    kwargs["reuse_weights"] = True
            else:
                weights_filename = None

            # Extract args that are to be given at call time.
            # output_chunks is only valid for xesmf >= 0.8, so don't add it be default to the call_kwargs
            call_kwargs = {"skipna": kwargs.pop("skipna", False)}
            if "output_chunks" in kwargs:
                call_kwargs["output_chunks"] = kwargs.pop("output_chunks")

            regridder = _regridder(
                ds_in=ds, ds_grid=ds_grid, filename=weights_filename, **kwargs
            )

            # The regridder (when fed Datasets) doesn't like if 'mask' is present.
            if "mask" in ds:
                ds = ds.drop_vars(["mask"])

            out = regridder(ds, keep_attrs=True, **call_kwargs)

            # Double-check that grid_mapping information is transferred
            gridmap_out = get_grid_mapping(ds_grid)
            if gridmap_out:
                # Regridder seems to seriously mess up the rotated dimensions
                for d in out.lon.dims:
                    out[d] = ds_grid[d]
                    if d not in out.coords:
                        out = out.assign_coords({d: ds_grid[d]})
                # Add the grid_mapping attribute
                for v in out.data_vars:
                    if any(
                        d in out[v].dims
                        for d in [out.cf.axes["X"][0], out.cf.axes["Y"][0]]
                    ):
                        out[v].attrs["grid_mapping"] = gridmap_out
                # Add the grid_mapping coordinate
                if gridmap_out not in out:
                    out = out.assign_coords({gridmap_out: ds_grid[gridmap_out]})
            else:
                gridmap_in = get_grid_mapping(ds)
                # Remove the original grid_mapping attribute
                for v in out.data_vars:
                    if "grid_mapping" in out[v].attrs:
                        out[v].attrs.pop("grid_mapping")
                # Remove the original grid_mapping coordinate if it is still in the output
                out = out.drop_vars(gridmap_in, errors="ignore")

            # cf_grid_2d adds temporary variables that we don't want to keep
            if "lon_bounds" in out and has_lon_bounds is False:
                out = out.drop_vars("lon_bounds")
            if "lat_bounds" in out and has_lat_bounds is False:
                out = out.drop_vars("lat_bounds")

            # History
            kwargs_for_hist = deepcopy(regridder_kwargs)
            kwargs_for_hist.setdefault("method", regridder.method)
            if intermediate_grids and i < len(intermediate_grids):
                name_inter = list(intermediate_grids.keys())[i]
                cf_grid_2d_args = intermediate_grids[name_inter]["cf_grid_2d"]
                new_history = (
                    f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                    f"regridded with regridder arguments {kwargs_for_hist} to a xesmf"
                    f" cf_grid_2d with arguments {cf_grid_2d_args}  - xESMF v{xe.__version__}"
                )
            else:
                new_history = (
                    f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                    f"regridded with arguments {kwargs_for_hist} - xESMF v{xe.__version__}"
                )
            history = (
                f"{new_history}\n{out.attrs['history']}"
                if "history" in out.attrs
                else new_history
            )
            out.attrs["history"] = history

    out = out.drop_vars("latitude_longitude", errors="ignore")
    # Attrs
    out.attrs["cat:processing_level"] = to_level
    out.attrs["cat:domain"] = (
        ds_grid.attrs["cat:domain"] if "cat:domain" in ds_grid.attrs else None
    )
    return out


@parse_config
def create_mask(
    ds: xr.Dataset | xr.DataArray,
    *,
    variable: str | None = None,
    where_operator: str | None = None,
    where_threshold: float | str | None = None,
    mask_nans: bool = True,
) -> xr.DataArray:
    """Create a 0-1 mask based on incoming arguments.

    Parameters
    ----------
    ds : xr.Dataset or xr.DataArray
        Dataset or DataArray to be evaluated. If a time dimension is present, the first time step will be used.
    variable : str, optional
        If using a Dataset, the variable on which to base the mask.
    where_operator : str, optional
        Operator to use for the threshold comparison. One of "<", "<=", "==", "!=", ">=", ">".
        Needs to be used with `where_threshold`.
    where_threshold : float or str, optional
        Threshold value to use for the comparison. A string can be used to reference units, e.g. "10 mm/day".
        Needs to be used with `where_operator`.
    mask_nans : bool, optional
        Whether to mask NaN values in the mask array. Default is True.

    Returns
    -------
    xr.DataArray
        Mask array.
    """
    # Prepare the mask for the destination grid
    ops = {
        "<": operator.lt,
        "<=": operator.le,
        "==": operator.eq,
        "!=": operator.ne,
        ">=": operator.ge,
        ">": operator.gt,
    }

    def cmp(arg1, op, arg2):
        operation = ops.get(op)
        return operation(arg1, arg2)

    if isinstance(ds, xr.Dataset):
        if variable is None:
            raise ValueError("A variable needs to be specified when passing a Dataset.")
        ds = ds[variable].copy()
    else:
        ds = ds.copy()
    if "time" in ds.dims:
        ds = ds.isel(time=0)

    mask = xr.ones_like(ds)
    mask.attrs = {"long_name": "Mask"}
    mask.name = "mask"

    # Create the mask based on the threshold
    if (where_operator is not None and where_threshold is None) or (
        where_operator is None and where_threshold is not None
    ):
        raise ValueError(
            "'where_operator' and 'where_threshold' must be used together."
        )
    if where_threshold is not None:
        mask.attrs["where_threshold"] = f"{variable} {where_operator} {where_threshold}"
        if isinstance(where_threshold, str):
            ds = convert_units_to(ds, where_threshold.split(" ")[1])
            where_threshold = float(where_threshold.split(" ")[0])

        mask = xr.where(
            cmp(ds, where_operator, where_threshold), mask, 0, keep_attrs=True
        )

    # Mask NaNs
    if mask_nans:
        mask = xr.where(ds.notnull(), mask, 0, keep_attrs=True)
        mask.attrs["mask_NaNs"] = "True"
    else:
        # The where clause above will mask NaNs, so we need to revert that
        attrs = mask.attrs
        mask = xr.where(ds.isnull(), 1, mask)
        mask.attrs = attrs
        mask.attrs["mask_NaNs"] = "False"

    return mask


def _regridder(
    ds_in: xr.Dataset,
    ds_grid: xr.Dataset,
    *,
    filename: str | os.PathLike | None = None,
    method: str = "bilinear",
    unmapped_to_nan: bool | None = True,
    **kwargs,
) -> Regridder:
    """Call to xesmf Regridder with a few default arguments.

    Parameters
    ----------
    ds_in : xr.Dataset
        Incoming grid. The Dataset needs to have lat/lon coordinates.
    ds_grid : xr.Dataset
        Destination grid. The Dataset needs to have lat/lon coordinates.
    filename : str or os.PathLike, optional
        Path to the NetCDF file with weights information.
    method : str
        Interpolation method.
    unmapped_to_nan : bool, optional
        Arguments to send xe.Regridder().
    regridder_kwargs : dict
        Arguments to send xe.Regridder().

    Returns
    -------
    xe.frontend.Regridder
        Regridder object
    """
    if method.startswith("conservative"):
        gridmap_in = get_grid_mapping(ds_in)
        gridmap_grid = get_grid_mapping(ds_grid)

        if (
            ds_in.cf["longitude"].ndim == 2
            and "longitude" not in ds_in.cf.bounds
            and gridmap_in in ds_in
        ):
            ds_in = ds_in.update(create_bounds_gridmapping(ds_in, gridmap_in))
        if (
            ds_grid.cf["longitude"].ndim == 2
            and "longitude" not in ds_grid.cf.bounds
            and gridmap_grid in ds_grid
        ):
            ds_grid = ds_grid.update(create_bounds_gridmapping(ds_grid, gridmap_grid))

    regridder = xe.Regridder(
        ds_in=ds_in,
        ds_out=ds_grid,
        method=method,
        unmapped_to_nan=unmapped_to_nan,
        **kwargs,
    )
    if filename is not None and not Path(filename).is_file():
        regridder.to_netcdf(filename)

    return regridder


def create_bounds_rotated_pole(ds: xr.Dataset) -> xr.Dataset:
    warnings.warn(
        "This function is deprecated and will be removed in xscen v0.12.0. Use create_bounds_gridmapping instead.",
        FutureWarning,
    )
    return create_bounds_gridmapping(ds, "rotated_pole")


def create_bounds_gridmapping(ds: xr.Dataset, gridmap: str | None = None) -> xr.Dataset:
    """Create bounds for rotated pole datasets."""
    if gridmap is None:
        gridmap = get_grid_mapping(ds)
        if gridmap == "":
            raise ValueError("Grid mapping could not be inferred from the dataset.")

    xname = ds.cf.axes["X"][0]
    yname = ds.cf.axes["Y"][0]

    ds = ds.cf.add_bounds([yname, xname])

    # In "vertices" format then expand to 2D. From (N, 2) to (N+1,) to (N+1, M+1)
    yv1D = cfxr.bounds_to_vertices(ds[f"{yname}_bounds"], "bounds")
    xv1D = cfxr.bounds_to_vertices(ds[f"{xname}_bounds"], "bounds")
    yv = yv1D.expand_dims(dict([(f"{xname}_vertices", xv1D)])).transpose(
        f"{xname}_vertices", f"{yname}_vertices"
    )
    xv = xv1D.expand_dims(dict([(f"{yname}_vertices", yv1D)])).transpose(
        f"{xname}_vertices", f"{yname}_vertices"
    )

    # Some CRS have additional attributes according to CF conventions
    def _get_opt_attr_as_float(da: xr.DataArray, attr: str) -> float | None:
        return float(da.attrs[attr]) if attr in da.attrs else None

    if gridmap == "rotated_pole":
        # Get cartopy's crs for the projection
        RP = ccrs.RotatedPole(
            pole_longitude=float(ds.rotated_pole.grid_north_pole_longitude),
            pole_latitude=float(ds.rotated_pole.grid_north_pole_latitude),
            central_rotated_longitude=_get_opt_attr_as_float(
                ds.rotated_pole, "north_pole_grid_longitude"
            ),
        )
    elif gridmap == "oblique_mercator":
        RP = ccrs.ObliqueMercator(
            central_longitude=float(ds.oblique_mercator.longitude_of_projection_origin),
            central_latitude=float(ds.oblique_mercator.latitude_of_projection_origin),
            false_easting=_get_opt_attr_as_float(ds.oblique_mercator, "false_easting"),
            false_northing=_get_opt_attr_as_float(
                ds.oblique_mercator, "false_northing"
            ),
            scale_factor=float(ds.oblique_mercator.scale_factor_at_projection_origin),
            azimuth=float(ds.oblique_mercator.azimuth_of_central_line),
        )
    else:
        raise NotImplementedError(f"Grid mapping {gridmap} not yet implemented.")

    PC = ccrs.PlateCarree()

    # Project points
    pts = PC.transform_points(RP, xv.values, yv.values)
    lonv = xv.copy(data=pts[..., 0]).rename("lon_vertices")
    latv = yv.copy(data=pts[..., 1]).rename("lat_vertices")

    # Back to CF bounds format. From (N+1, M+1) to (4, N, M)
    lonb = cfxr.vertices_to_bounds(lonv, ("bounds", xname, yname)).rename("lon_bounds")
    latb = cfxr.vertices_to_bounds(latv, ("bounds", xname, yname)).rename("lat_bounds")

    # Create dataset, set coords and attrs
    ds_bnds = xr.merge([lonb, latb]).assign(
        dict([("lon", ds.lon), ("lat", ds.lat), (gridmap, ds[gridmap])])
    )
    ds_bnds[yname] = ds[yname]
    ds_bnds[xname] = ds[xname]
    ds_bnds.lat.attrs["bounds"] = "lat_bounds"
    ds_bnds.lon.attrs["bounds"] = "lon_bounds"
    return ds_bnds.transpose(*ds.lon.dims, "bounds")


def _generate_random_string(length: int):
    characters = string.ascii_letters + string.digits

    # Random seed based on the current time
    random.seed(datetime.datetime.now().timestamp())
    random_string = "".join(
        random.choice(characters) for i in range(length)  # noqa: S311
    )
    return random_string



"""Configuration module.

Configuration in this module is taken from yaml files.

Functions wrapped by :py:func:`parse_config` have their kwargs automatically patched by
values in the config.

The ``CONFIG`` dictionary contains all values, structured by submodules and functions. For example,
for function ``function`` defined in ``module.py`` of this package, the config would look like:

.. code-block:: yaml

    module:
        function:
            ...kwargs...


The :py:func:`load_config` function fills the ``CONFIG`` dict from yaml files.
It always updates the dictionary, so the latest file read has the highest priority.

At calling time, the priority order is always (from highest to lowest priority):

1. Explicitly passed keyword-args
2. Values in the loaded config
3. Function's default values.

Special sections
~~~~~~~~~~~~~~~~
After parsing the files, :py:func:`load_config` will look into the config and perform some
extra actions when finding the following special sections:

- ``logging``:
  The content of this section will be sent directly to :py:func:`logging.config.dictConfig`.
- ``xarray``:
  The content of this section will be sent directly to :py:func:`xarray.set_options`.
- ``xclim``:
  The content of this section will be sent directly to :py:func:`xclim.set_options`.
  Here goes `metadata_locales: - fr` to activate the automatic translation of added attributes, for example.
- ``warnings``:
  The content of this section must be a simple mapping. The keys are understood as python
  warning categories (types) and the values as an action to add to the filter. The key "all"
  applies the filter to any warnings. Only built-in warnings are supported.
"""

import ast
import builtins
import collections.abc
import inspect
import logging.config
import types
import warnings
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Any

import xarray as xr
import xclim as xc
import yaml

logger = logging.getLogger(__name__)
EXTERNAL_MODULES = ["logging", "xarray", "xclim", "warnings"]

__all__ = [
    "CONFIG",
    "args_as_str",
    "load_config",
    "parse_config",
    "recursive_update",
]


class ConfigDict(dict):
    """A special dictionary that returns a copy on getitem."""

    def __getitem__(self, key):
        value = super().__getitem__(key)
        if isinstance(value, collections.abc.Mapping):
            return ConfigDict(deepcopy(value))
        return value

    def set(self, key, value):
        parts = key.split(".")
        d = self
        for part in parts[:-1]:
            d = d.setdefault(part, {})
            if not isinstance(d, collections.abc.Mapping):
                raise ValueError(
                    f"Key {key} points to an invalid config section ({part} if not a mapping)."
                )
        d[parts[-1]] = value

    def update_from_list(self, pairs):
        for key, valstr in pairs:
            try:
                val = ast.literal_eval(valstr)
            except (SyntaxError, ValueError):
                val = valstr
            self.set(key, val)


CONFIG = ConfigDict()


def recursive_update(d, other):
    """Update a dictionary recursively with another dictionary.

    Values that are Mappings are updated recursively as well.
    """
    for k, v in other.items():
        if isinstance(v, collections.abc.Mapping):
            old_v = d.get(k)
            if isinstance(old_v, collections.abc.Mapping):
                d[k] = recursive_update(old_v, v)
            else:
                d[k] = v
        else:
            d[k] = v
    return d


def args_as_str(*args: tuple[Any, ...]) -> tuple[str, ...]:
    """Return arguments as strings."""
    new_args = []
    for i, arg in enumerate(*args):
        if isinstance(arg, Path):
            new_args.append(str(arg))
        else:
            new_args.append(arg)
    return tuple(new_args)


def load_config(
    *elements,
    reset: bool = False,
    encoding: str | None = None,
    verbose: bool = False,
):
    """Load configuration from given files or key=value pairs.

    Once all elements are loaded, special sections are dispatched to their module, but only if
    the section was changed by the loaded elements. These special sections are:

    * `locales` : The locales to use when writing metadata in xscen, xclim and figanos. This section must be a list of 2-char strings.
    * `logging` : Everything passed to :py:func:`logging.config.dictConfig`.
    * `xarray` : Passed to :py:func:`xarray.set_options`.
    * `xclim` : Passed to :py:func:`xclim.set_options`.
    * `warning` : Mappings where the key is a Warning category (or "all") and the value an action to pass to :py:func:`warnings.simplefilter`.

    Parameters
    ----------
    elements : str
        Files or values to add into the config.
        If a directory is passed, all `.yml` files of this directory are added, in alphabetical order.
        If a "key=value" string, "key" is a dotted name and value will be evaluated if possible.
        "key=value" pairs are set last, after all files are being processed.
    reset : bool
        If True, erases the current config before loading files.
    encoding : str, optional
        The encoding to use when reading files.
    verbose: bool
        If True, each element triggers a INFO log line.

    Example
    -------
    .. code-block:: python

        load_config("my_config.yml", "config_dir/", "logging.loggers.xscen.level=DEBUG")

    Will load configuration from `my_config.yml`, then from all yml files in `config_dir`
    and then the logging level of xscen's logger will be set to DEBUG.
    """
    if reset:
        CONFIG.clear()

    old_external = [deepcopy(CONFIG.get(module, {})) for module in EXTERNAL_MODULES]

    # Use of map(Path, ...) ensures that "file" is a Path, no matter if a Path or a str was given.
    for element in elements:
        if "=" in element:
            key, value = element.split("=")
            CONFIG.update_from_list([(key, value)])
            if verbose:
                msg = f"Updated the config with {element}."
                logger.info(msg)
        else:
            file = Path(element)
            if file.is_dir():
                # Get all yml files, sort by name
                configfiles = sorted(file.glob("*.yml"), key=lambda p: p.name)
            else:
                configfiles = [file]

            for configfile in configfiles:
                with configfile.open(encoding=encoding) as f:
                    recursive_update(CONFIG, yaml.safe_load(f))
                    if verbose:
                        msg = f"Updated the config with {configfile}."
                        logger.info(msg)

    for module, old in zip(EXTERNAL_MODULES, old_external):
        if old != CONFIG.get(module, {}):
            _setup_external(module, CONFIG.get(module, {}))


def parse_config(func_or_cls):  # noqa: D103
    module = ".".join(func_or_cls.__module__.split(".")[1:])

    if isinstance(func_or_cls, type):
        func = func_or_cls.__init__
    else:
        func = func_or_cls

    @wraps(func)
    def _wrapper(*args, **kwargs):
        # Get dotted module name, excluding the main package name.

        from_config = CONFIG.get(module, {}).get(func.__name__, {})
        sig = inspect.signature(func)
        if CONFIG.get("print_it_all"):
            msg = f"For func {func}, found config {from_config}.\nOriginal kwargs : {kwargs}"
            logger.debug(msg)
        for k, v in from_config.items():
            if k in sig.parameters:
                kwargs.setdefault(k, v)
        if CONFIG.get("print_it_all"):
            msg = f"Modified kwargs : {kwargs}"
            logger.debug(msg)

        return func(*args, **kwargs)

    _wrapper.configurable = True
    if isinstance(func_or_cls, type):
        func_or_cls.__init__ = _wrapper
        return func_or_cls
    _wrapper: func_or_cls  # Fix a decorator bug in Pycharm 2022
    return _wrapper


def _setup_external(module, config):
    if module == "logging":
        config.update(version=1)
        logging.config.dictConfig(config)
    elif module == "xclim":
        xc.set_options(**config)
    elif module == "xarray":
        xr.set_options(**config)
    elif module == "warning":
        for category, action in config.items():
            if category == "all":
                warnings.simplefilter(action)
            elif issubclass(getattr(builtins, category), builtins.Warning):
                warnings.simplefilter(action, category=getattr(builtins, category))


def get_configurable():
    """Return a dictionary of all configurable functions and classes of xscen."""
    import xscen as xs

    configurable = {}
    for module in dir(xs):
        modobj = getattr(xs, module)
        if isinstance(modobj, types.ModuleType):
            for func in dir(modobj):
                funcobj = getattr(modobj, func)
                if getattr(funcobj, "configurable", False) or getattr(
                    getattr(funcobj, "__init__", None), "configurable", False
                ):
                    configurable[f"xscen.{module}.{func}"] = funcobj
    return configurable



"""Spatial tools."""

import datetime
import itertools
import logging
import warnings
from collections.abc import Sequence
from pathlib import Path

import clisops.core.subset
import dask
import geopandas as gpd
import numpy as np
import sparse as sp
import xarray as xr
import xclim as xc
from pyproj.crs import CRS

from .config import parse_config

logger = logging.getLogger(__name__)

__all__ = [
    "creep_fill",
    "creep_weights",
    "get_grid_mapping",
    "subset",
]


@parse_config
def creep_weights(mask: xr.DataArray, n: int = 1, mode: str = "clip") -> xr.DataArray:
    """Compute weights for the creep fill.

    The output is a sparse matrix with the same dimensions as `mask`, twice.

    Parameters
    ----------
    mask : DataArray
      A boolean DataArray. False values are candidates to the filling.
      Usually they represent missing values (`mask = da.notnull()`).
      All dimensions are creep filled.
    n : int
      The order of neighbouring to use. 1 means only the adjacent grid cells are used.
    mode : {'clip', 'wrap'}
      If a cell is on the edge of the domain, `mode='wrap'` will wrap around to find neighbours.

    Returns
    -------
    DataArray
       Weights. The dot product must be taken over the last N dimensions.
    """
    if mode not in ["clip", "wrap"]:
        raise ValueError("mode must be either 'clip' or 'wrap'")
    da = mask
    mask = da.values
    neighbors = np.array(
        list(itertools.product(*[np.arange(-n, n + 1) for j in range(mask.ndim)]))
    ).T
    src = []
    dst = []
    w = []
    it = np.nditer(mask, flags=["f_index", "multi_index"], order="C")
    for i in it:
        if not i:
            neigh_idx_2d = np.atleast_2d(it.multi_index).T + neighbors
            neigh_idx_1d = np.ravel_multi_index(
                neigh_idx_2d, mask.shape, order="C", mode=mode
            )
            if mode == "clip":
                neigh_idx = np.unravel_index(
                    np.unique(neigh_idx_1d), mask.shape, order="C"
                )
            elif mode == "wrap":
                neigh_idx = np.unravel_index(neigh_idx_1d, mask.shape, order="C")
            neigh = mask[neigh_idx]
            N = (neigh).sum()
            if N > 0:
                src.extend([it.multi_index] * N)
                dst.extend(np.stack(neigh_idx)[:, neigh].T)
                w.extend([1 / N] * N)
            else:
                src.extend([it.multi_index])
                dst.extend([it.multi_index])
                w.extend([np.nan])
        else:
            src.extend([it.multi_index])
            dst.extend([it.multi_index])
            w.extend([1])
    crds = np.concatenate((np.array(src).T, np.array(dst).T), axis=0)
    return xr.DataArray(
        sp.COO(crds, w, (*da.shape, *da.shape)),
        dims=[f"{d}_out" for d in da.dims] + list(da.dims),
        coords=da.coords,
        name="creep_fill_weights",
    )


@parse_config
def creep_fill(da: xr.DataArray, w: xr.DataArray) -> xr.DataArray:
    """Creep fill using pre-computed weights.

    Parameters
    ----------
    da: DataArray
      A DataArray sharing the dimensions with the one used to compute the weights.
      It can have other dimensions.
      Dask is supported as long as there are no chunks over the creeped dims.
    w: DataArray
      The result of `creep_weights`.

    Returns
    -------
    xarray.DataArray, same shape as `da`, but values filled according to `w`.

    Examples
    --------
    >>> w = creep_weights(da.isel(time=0).notnull(), n=1)
    >>> da_filled = creep_fill(da, w)
    """

    def _dot(arr, wei):
        N = wei.ndim // 2
        extra_dim = arr.ndim - N
        return np.tensordot(arr, wei, axes=(np.arange(N) + extra_dim, np.arange(N) + N))

    N = w.ndim // 2
    return xr.apply_ufunc(
        _dot,
        da,
        w,
        input_core_dims=[w.dims[N:], w.dims],
        output_core_dims=[w.dims[N:]],
        dask="parallelized",
        output_dtypes=["float64"],
    )


def subset(
    ds: xr.Dataset,
    method: str,
    *,
    name: str | None = None,
    tile_buffer: float = 0,
    **kwargs,
) -> xr.Dataset:
    r"""
    Subset the data to a region.

    Either creates a slice and uses the .sel() method, or customizes a call to
    clisops.subset() that allows for an automatic buffer around the region.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be subsetted.
    method : {'gridpoint', 'bbox', shape', 'sel'}
        If the method is 'sel', this is not a call to clisops but only a subsetting with the xarray `.sel()` function.
    name : str, optional
        Used to rename the 'cat:domain' attribute.
    tile_buffer : float
        For ['bbox', shape'], uses an approximation of the grid cell size to add a buffer around the requested region.
        This differs from clisops' 'buffer' argument in subset_shape().
    \*\*kwargs : dict
        Arguments to be sent to clisops. See relevant function for details. Depending on the method, required kwargs are:
        - gridpoint: lon, lat
        - bbox: lon_bnds, lat_bnds
        - shape: shape
        - sel: slices for each dimension

    Returns
    -------
    xr.Dataset
        Subsetted Dataset.

    See Also
    --------
    clisops.core.subset.subset_gridpoint, clisops.core.subset.subset_bbox, clisops.core.subset.subset_shape
    """
    if tile_buffer > 0 and method in ["gridpoint", "sel"]:
        warnings.warn(
            f"tile_buffer is not used for the '{method}' method. Ignoring the argument.",
            UserWarning,
        )

    if "latitude" not in ds.cf or "longitude" not in ds.cf:
        ds = ds.cf.guess_coord_axis()

    if method == "gridpoint":
        ds_subset = _subset_gridpoint(ds, name=name, **kwargs)
    elif method == "bbox":
        ds_subset = _subset_bbox(ds, name=name, tile_buffer=tile_buffer, **kwargs)
    elif method == "shape":
        ds_subset = _subset_shape(ds, name=name, tile_buffer=tile_buffer, **kwargs)
    elif method == "sel":
        ds_subset = _subset_sel(ds, name=name, **kwargs)
    else:
        raise ValueError(
            "Subsetting type not recognized. Use 'gridpoint', 'bbox', 'shape' or 'sel'."
        )

    return ds_subset


def _subset_gridpoint(
    ds: xr.Dataset,
    lon: float | Sequence[float] | xr.DataArray,
    lat: float | Sequence[float] | xr.DataArray,
    *,
    name: str | None = None,
    **kwargs,
) -> xr.Dataset:
    r"""Subset the data to a gridpoint.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be subsetted.
    lon : float or Sequence[float] or xr.DataArray
        Longitude coordinate(s). Must be of the same length as lat.
    lat : float or Sequence[float] or xr.DataArray
        Latitude coordinate(s). Must be of the same length as lon.
    name: str, optional
        Used to rename the 'cat:domain' attribute.
    \*\*kwargs : dict
        Other arguments to be sent to clisops. Possible kwargs are:
        - start_date (str): Start date for the subset in the format 'YYYY-MM-DD'.
        - end_date (str): End date for the subset in the format 'YYYY-MM-DD'.
        - first_level (int or float): First level of the subset.
        - last_level (int or float): Last level of the subset.
        - tolerance (float): Masks values if the distance to the nearest gridpoint is larger than tolerance in meters.
        - add_distance (bool): If True, adds a variable with the distance to the nearest gridpoint.

    Returns
    -------
    xr.Dataset
        Subsetted Dataset.
    """
    ds = _load_lon_lat(ds)
    if not hasattr(lon, "__iter__"):
        lon = [lon]
    if not hasattr(lat, "__iter__"):
        lat = [lat]

    ds_subset = clisops.core.subset_gridpoint(ds, lon=lon, lat=lat, **kwargs)
    new_history = (
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
        f"gridpoint spatial subsetting on {len(lon)} coordinates - clisops v{clisops.__version__}"
    )

    return update_history_and_name(ds_subset, new_history, name)


def _subset_bbox(
    ds: xr.Dataset,
    lon_bnds: tuple[float, float] | list[float],
    lat_bnds: tuple[float, float] | list[float],
    *,
    name: str | None = None,
    tile_buffer: float = 0,
    **kwargs,
) -> xr.Dataset:
    r"""Subset the data to a bounding box.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be subsetted.
    lon_bnds : tuple or list of two floats
        Longitude boundaries of the bounding box.
    lat_bnds : tuple or list of two floats
        Latitude boundaries of the bounding box.
    name: str, optional
        Used to rename the 'cat:domain' attribute.
    tile_buffer: float
        Uses an approximation of the grid cell size to add a dynamic buffer around the requested region.
    \*\*kwargs : dict
        Other arguments to be sent to clisops. Possible kwargs are:
        - start_date (str): Start date for the subset in the format 'YYYY-MM-DD'.
        - end_date (str): End date for the subset in the format 'YYYY-MM-DD'.
        - first_level (int or float): First level of the subset.
        - last_level (int or float): Last level of the subset.
        - time_values (Sequence[str]): A list of datetime strings to subset.
        - level_values (Sequence[int or float]): A list of levels to subset.

    Returns
    -------
    xr.Dataset
        Subsetted Dataset.
    """
    ds = _load_lon_lat(ds)

    if tile_buffer > 0:
        lon_res, lat_res = _estimate_grid_resolution(ds)
        lon_bnds = (
            lon_bnds[0] - lon_res * tile_buffer,
            lon_bnds[1] + lon_res * tile_buffer,
        )
        lat_bnds = (
            lat_bnds[0] - lat_res * tile_buffer,
            lat_bnds[1] + lat_res * tile_buffer,
        )

    ds_subset = clisops.core.subset_bbox(
        ds, lon_bnds=lon_bnds, lat_bnds=lat_bnds, **kwargs
    )
    new_history = (
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
        f"bbox spatial subsetting with {'buffer=' + str(tile_buffer) if tile_buffer > 0 else 'no buffer'}"
        f", lon_bnds={np.array(lon_bnds)}, lat_bnds={np.array(lat_bnds)}"
        f" - clisops v{clisops.__version__}"
    )

    return update_history_and_name(ds_subset, new_history, name)


def _subset_shape(
    ds: xr.Dataset,
    shape: str | Path | gpd.GeoDataFrame,
    *,
    name: str | None = None,
    tile_buffer: float = 0,
    **kwargs,
) -> xr.Dataset:
    r"""Subset the data to a shape.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be subsetted.
    shape : str or gpd.GeoDataFrame
        Path to the shapefile or GeoDataFrame.
    name: str, optional
        Used to rename the 'cat:domain' attribute.
    tile_buffer: float
        Uses an approximation of the grid cell size to add a buffer around the requested region.
    \*\*kwargs : dict
        Other arguments to be sent to clisops. Possible kwargs are:
        - raster_crs (str or int): EPSG number or PROJ4 string.
        - shape_crs (str or int): EPSG number or PROJ4 string.
        - buffer (float): Buffer size to add around the shape. Units are based on the shape degrees/metres.
        - start_date (str): Start date for the subset in the format 'YYYY-MM-DD'.
        - end_date (str): End date for the subset in the format 'YYYY-MM-DD'.
        - first_level (int or float): First level of the subset.
        - last_level (int or float): Last level of the subset.

    Returns
    -------
    xr.Dataset
        Subsetted Dataset.
    """
    ds = _load_lon_lat(ds)

    if tile_buffer > 0:
        if kwargs.get("buffer") is not None:
            raise ValueError(
                "Both tile_buffer and clisops' buffer were requested. Use only one."
            )
        lon_res, lat_res = _estimate_grid_resolution(ds)

        # The buffer argument needs to be in the same units as the shapefile, so it's simpler to always project the shapefile to WGS84.
        if isinstance(shape, str | Path):
            shape = gpd.read_file(shape)

        try:
            shape_crs = shape.crs
        except AttributeError:
            shape_crs = None
        if shape_crs is None:
            warnings.warn(
                "Shapefile does not have a CRS. Compatibility with the dataset is not guaranteed.",
                category=UserWarning,
            )
        elif shape_crs != CRS(4326):  # WGS84
            warnings.warn(
                "Shapefile is not in EPSG:4326. Reprojecting to this CRS.",
                UserWarning,
            )
            shape = shape.to_crs(4326)

        kwargs["buffer"] = np.max([lon_res, lat_res]) * tile_buffer

    ds_subset = clisops.core.subset_shape(ds, shape=shape, **kwargs)
    new_history = (
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
        f"shape spatial subsetting with {'buffer=' + str(tile_buffer) if tile_buffer > 0 else 'no buffer'}"
        f", shape={Path(shape).name if isinstance(shape, str | Path) else 'gpd.GeoDataFrame'}"
        f" - clisops v{clisops.__version__}"
    )

    return update_history_and_name(ds_subset, new_history, name)


def _subset_sel(ds: xr.Dataset, *, name: str | None = None, **kwargs) -> xr.Dataset:
    r"""Subset the data using the .sel() method.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to be subsetted.
    name: str, optional
        Used to rename the 'cat:domain' attribute.
    \*\*kwargs : dict
        The keys are the dimensions to subset and the values are turned into a slice.

    Returns
    -------
    xr.Dataset
        Subsetted Dataset.
    """
    # Create a dictionary with slices for each dimension
    arg_sel = {dim: slice(*map(float, bounds)) for dim, bounds in kwargs.items()}

    # Subset the dataset
    ds_subset = ds.sel(**arg_sel)

    # Update the history attribute
    new_history = (
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
        f"sel subsetting with arguments {arg_sel}"
    )

    return update_history_and_name(ds_subset, new_history, name)


def _load_lon_lat(ds: xr.Dataset) -> xr.Dataset:
    """Load longitude and latitude for more efficient subsetting."""
    if xc.core.utils.uses_dask(ds.cf["longitude"]):
        logger.info("Loading longitude for more efficient subsetting.")
        (ds[ds.cf["longitude"].name],) = dask.compute(ds[ds.cf["longitude"].name])
    if xc.core.utils.uses_dask(ds.cf["latitude"]):
        logger.info("Loading latitude for more efficient subsetting.")
        (ds[ds.cf["latitude"].name],) = dask.compute(ds[ds.cf["latitude"].name])

    return ds


def get_grid_mapping(ds: xr.Dataset) -> str:
    """Get the grid_mapping attribute from the dataset."""
    gridmap = [
        ds[v].attrs["grid_mapping"]
        for v in ds.data_vars
        if "grid_mapping" in ds[v].attrs
    ]
    gridmap += [c for c in ds.coords if ds[c].attrs.get("grid_mapping_name", None)]
    gridmap = list(np.unique(gridmap))

    if len(gridmap) > 1:
        warnings.warn(
            f"There are conflicting grid_mapping attributes in the dataset. Assuming {gridmap[0]}."
        )

    return gridmap[0] if gridmap else ""


def _estimate_grid_resolution(ds: xr.Dataset) -> tuple[float, float]:
    # Since this is to compute a buffer, we take the maximum difference as an approximation.
    # Estimate the grid resolution
    if len(ds.lon.dims) == 1:  # 1D lat-lon
        lon_res = np.abs(ds.lon.diff("lon").max().values)
        lat_res = np.abs(ds.lat.diff("lat").max().values)
    else:
        lon_res = np.abs(ds.lon.diff(ds.cf["X"].name).max().values)
        lat_res = np.abs(ds.lat.diff(ds.cf["Y"].name).max().values)

    return lon_res, lat_res


def update_history_and_name(ds_subset, new_history, name):
    history = (
        new_history + " \n " + ds_subset.attrs["history"]
        if "history" in ds_subset.attrs
        else new_history
    )
    ds_subset.attrs["history"] = history
    if name is not None:
        ds_subset.attrs["cat:domain"] = name
    return ds_subset



"""Common utilities to be used in many places."""

import fnmatch
import gettext
import json
import logging
import os
import re
import warnings
from collections import defaultdict
from collections.abc import Sequence
from copy import deepcopy
from datetime import datetime
from itertools import chain
from pathlib import Path
from types import ModuleType

import cftime
import flox.xarray
import numpy as np
import pandas as pd
import xarray as xr
from xarray.coding import cftime_offsets as cfoff
from xclim.core import units
from xclim.core.calendar import parse_offset
from xclim.core.options import METADATA_LOCALES
from xclim.core.options import OPTIONS as XC_OPTIONS
from xclim.core.utils import uses_dask

from .config import parse_config

logger = logging.getLogger(__name__)

__all__ = [
    "CV",
    "add_attr",
    "change_units",
    "clean_up",
    "date_parser",
    "ensure_correct_time",
    "ensure_new_xrfreq",
    "get_cat_attrs",
    "maybe_unstack",
    "minimum_calendar",
    "natural_sort",
    "stack_drop_nans",
    "standardize_periods",
    "translate_time_chunk",
    "unstack_dates",
    "unstack_fill_nan",
    "update_attr",
    "xrfreq_to_timedelta",
]

TRANSLATOR = defaultdict(lambda: lambda s: s)
"""Dictionary of translating objects.

Each key is a two letter locale code and values are functions that return the translated message
as compiled in the gettext catalogs. If a language is not defined or a message not translated,
the function will return the raw message.
"""

try:
    for loc in (Path(__file__).parent / "data").iterdir():
        if loc.is_dir() and len(loc.name) == 2:
            TRANSLATOR[loc.name] = gettext.translation(
                "xscen", localedir=loc.parent, languages=[loc.name]
            ).gettext
except FileNotFoundError as err:
    raise ImportError(
        "Your xscen installation doesn't have compiled translations. "
        "Run `make translate` from the source directory to fix."
    ) from err


def update_attr(
    ds: xr.Dataset | xr.DataArray,
    attr: str,
    new: str,
    others: Sequence[xr.Dataset | xr.DataArray] | None = None,
    **fmt,
) -> xr.Dataset | xr.DataArray:
    r"""Format an attribute referencing itself in a translatable way.

    Parameters
    ----------
    ds : Dataset or DataArray
        The input object with the attribute to update.
    attr : str
        Attribute name.
    new : str
        New attribute as a template string. It may refer to the old version
        of the attribute with the "{attr}" field.
    others: Sequence of Datasets or DataArrays
        Other objects from which we can extract the attribute `attr`.
        These can be referenced as "{attrXX}" in `new`, where XX is the based-1 index of the other source in `others`.
        If they don't have the `attr` attribute, an empty string is sent to the string formatting.
        See notes.
    \*\*fmt
        Other formatting data.

    Returns
    -------
    `ds`, but updated with the new version of `attr`, in each of the activated languages.

    Notes
    -----
    This is meant for constructing attributes by extending a previous version
    or combining it from different sources. For example, given a `ds` that has `long_name="Variability"`:

    >>> update_attr(ds, "long_name", _("Mean of {attr}"))

    Will update the "long_name" of `ds` with `long_name="Mean of Variability"`.
    The use of `_(...)` allows the detection of this string by the translation manager. The function
    will be able to add a translatable version of the string for each activated language, for example adding
    a `long_name_fr="Moyenne de Variabilité"` (assuming a `long_name_fr` was present on the initial `ds`).

    If the new attribute is an aggregation from multiple sources, these can be passed in `others`.

    >>> update_attr(
    ...     ds0,
    ...     "long_name",
    ...     _("Addition of {attr} and {attr1}, divided by {attr2}"),
    ...     others=[ds1, ds2],
    ... )

    Here, `ds0` will have it's `long_name` updated with the passed string, where  `attr1` is the `long_name` of `ds1`
    and `attr2` the `long_name` of `ds2`. The process will be repeated for each localized `long_name` available on `ds0`.
    For example, if `ds0` has a `long_name_fr`, the template string is translated and
    filled with the `long_name_fr` attributes of  `ds0`, `ds1` and `ds2`.
    If the latter don't exist, the english version is used instead.
    """
    others = others or []
    # .strip(' .') removes trailing and leading whitespaces and dots
    if attr in ds.attrs:

        others_attrs = {
            f"attr{i}": dso.attrs.get(attr, "").strip(" .")
            for i, dso in enumerate(others, 1)
        }
        ds.attrs[attr] = new.format(
            attr=ds.attrs[attr].strip(" ."), **others_attrs, **fmt
        )
    # All existing locales
    for key in fnmatch.filter(ds.attrs.keys(), f"{attr}_??"):
        loc = key[-2:]
        others_attrs = {
            f"attr{i}": dso.attrs.get(key, dso.attrs.get(attr, "")).strip(" .")
            for i, dso in enumerate(others, 1)
        }
        ds.attrs[key] = TRANSLATOR[loc](new).format(
            attr=ds.attrs[key].strip(" ."), **others_attrs, **fmt
        )


def add_attr(ds: xr.Dataset | xr.DataArray, attr: str, new: str, **fmt):
    """Add a formatted translatable attribute to a dataset."""
    ds.attrs[attr] = new.format(**fmt)
    for loc in XC_OPTIONS[METADATA_LOCALES]:
        ds.attrs[f"{attr}_{loc}"] = TRANSLATOR[loc](new).format(**fmt)


def date_parser(  # noqa: C901
    date: str | cftime.datetime | pd.Timestamp | datetime | pd.Period,
    *,
    end_of_period: bool | str = False,
    out_dtype: str = "datetime",
    strtime_format: str = "%Y-%m-%d",
    freq: str = "h",
) -> str | pd.Period | pd.Timestamp:
    """Return a datetime from a string.

    Parameters
    ----------
    date : str, cftime.datetime, pd.Timestamp, datetime.datetime, pd.Period
        Date to be converted
    end_of_period : bool or str
        If 'YE' or 'ME', the returned date will be the end of the year or month that contains the received date.
        If True, the period is inferred from the date's precision, but `date` must be a string,
        otherwise nothing is done.
    out_dtype : str
        Choices are 'datetime', 'period' or 'str'
    strtime_format : str
        If out_dtype=='str', this sets the strftime format
    freq : str
        If out_dtype=='period', this sets the frequency of the period.

    Returns
    -------
    pd.Timestamp, pd.Period, str
        Parsed date
    """
    # Formats, ordered depending on string length
    fmts = {
        4: ["%Y"],
        6: ["%Y%m"],
        7: ["%Y-%m"],
        8: ["%Y%m%d"],
        10: ["%Y%m%d%H", "%Y-%m-%d"],
        12: ["%Y%m%d%H%M"],
        16: ["%Y-%m-%d %H:%M"],
        19: ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S"],
    }

    def _parse_date(date, fmts):
        for fmt in fmts:
            try:
                # `pd.to_datetime` fails with out-of-bounds
                s = datetime.strptime(date, fmt)
            except ValueError:
                pass
            else:
                match = fmt
                break
        else:
            raise ValueError(f"Can't parse date {date} with formats {fmts}.")
        return s, match

    fmt = None
    # Timestamp can parse a few date formats by default, but not the ones without spaces
    # So we try a few known formats first, then a plain call
    # Also we need "fmt" to know the precision of the string (if end_of_period is True)
    if isinstance(date, str):
        try:
            date, fmt = _parse_date(date, fmts[len(date)])
        except (KeyError, ValueError):
            try:
                date = pd.Timestamp(date)
            except (ValueError, pd._libs.tslibs.parsing.DateParseError):
                date = pd.NaT
    elif isinstance(date, cftime.datetime):
        for n in range(3):
            try:
                date = pd.Timestamp((date - pd.Timedelta(n, "D")).isoformat())
            except ValueError:  # We are NOT catching OutOfBoundsDatetime.
                pass
            else:
                break
        else:
            raise ValueError(
                "Unable to parse cftime date {date}, even when moving back 2 days."
            )
    elif isinstance(date, pd.Period):
        # Pandas, you're a mess: Period.to_timestamp() fails for out-of-bounds dates (<1677, > 2242), but not when parsing a string...
        date = pd.Timestamp(date.strftime("%Y-%m-%dT%H:%M:%S"))

    if not isinstance(date, pd.Timestamp):
        date = pd.Timestamp(date)

    if isinstance(end_of_period, str) or (end_of_period is True and fmt):
        quasiday = (pd.Timedelta(1, "d") - pd.Timedelta(1, "s")).as_unit(date.unit)
        if end_of_period in ["Y", "YE"] or "m" not in fmt:
            date = (
                pd.tseries.frequencies.to_offset("YE-DEC").rollforward(date) + quasiday
            )
        elif end_of_period in ["M", "ME"] or "d" not in fmt:
            date = pd.tseries.frequencies.to_offset("ME").rollforward(date) + quasiday
        # TODO: Implement subdaily ?

    if out_dtype == "str":
        return date.strftime(strtime_format)
    elif out_dtype == "period":
        return date.to_period(freq)
    else:
        return date


def minimum_calendar(*calendars) -> str:
    """Return the minimum calendar from a list.

    Uses the hierarchy: 360_day < noleap < standard < all_leap, and returns one of those names.
    """
    # Unwrap any lists or tuples given in the input, but without destroying strings.
    calendars = [[cal] if isinstance(cal, str) else cal for cal in calendars]
    calendars = list(chain(*calendars))

    # Raise an error if the calendars are not recognized
    unknowns = set(calendars).difference(
        [
            "360_day",
            "365_day",
            "noleap",
            "standard",
            "default",
            "all_leap",
            "366_day",
            "gregorian",
            "proleptic_gregorian",
        ]
    )
    if unknowns:
        warnings.warn(
            f"These calendars are not recognized: {unknowns}. Results may be incorrect.",
        )

    if "360_day" in calendars:
        out = "360_day"
    elif "noleap" in calendars or "365_day" in calendars:
        out = "noleap"
    elif all(cal in ["all_leap", "366_day"] for cal in calendars):
        out = "all_leap"
    else:
        out = "standard"

    return out


def translate_time_chunk(chunks: dict, calendar: str, timesize: int) -> dict:
    """Translate chunk specification for time into a number.

    Parameters
    ----------
    chunks : dict
        Dictionary specifying the chunk sizes for each dimension. The time dimension can be specified as:
        -1 : translates to `timesize`
        'Nyear' : translates to N times the number of days in a year of the given calendar.
    calendar : str
        The calendar type (e.g., 'noleap', '360_day', 'all_leap').
    timesize : int
        The size of the time dimension.

    Returns
    -------
    dict
        The updated chunks dictionary with the time dimension translated to a number.

    Notes
    -----
    -1 translates to `timesize`
    'Nyear' translates to N times the number of days in a year of calendar `calendar`.
    """
    for k, v in chunks.items():
        if isinstance(v, dict):
            chunks[k] = translate_time_chunk(v.copy(), calendar, timesize)
        elif k == "time" and v is not None:
            if isinstance(v, str) and v.endswith("year"):
                n = int(chunks["time"].split("year")[0])
                nt = n * {
                    "noleap": 365,
                    "365_day": 365,
                    "360_day": 360,
                    "all_leap": 366,
                    "366_day": 366,
                }.get(calendar, 365.25)
                if nt != int(nt):
                    warnings.warn(
                        f"The number of days in {chunks['time']} for calendar {calendar} is not an integer. "
                        f"Chunks will not align perfectly with year ends."
                    )
                chunks[k] = int(nt)
            elif v == -1:
                chunks[k] = timesize
    return chunks


@parse_config
def stack_drop_nans(
    ds: xr.Dataset,
    mask: xr.DataArray | list[str],
    *,
    new_dim: str = "loc",
    to_file: str | None = None,
) -> xr.Dataset:
    """Stack dimensions into a single axis and drops indexes where the mask is false.

    Parameters
    ----------
    ds : xr.Dataset
      A dataset with the same coords as `mask`.
    mask : xr.DataArray or list of str
      A boolean DataArray with True on the points to keep. The mask will be loaded within this function, but not the dataset.
      Alternatively, a list of dimension names to stack. In this case, a mask will be created by loading all data and checking for NaNs.
      The latter is not recommended for large datasets.
    new_dim : str
      The name of the new stacked dim.
    to_file : str, optional
      A netCDF filename where to write the stacked coords for use in `unstack_fill_nan`.
      If given a string with {shape} and {domain}, the formatting will fill them with
      the original shape of the dataset and the global attributes 'cat:domain'.
      If None (default), nothing is written to disk.
      It is recommended to fill this argument in the config. It will be parsed automatically.
      E.g.:

          utils:
            stack_drop_nans:
                to_file: /some_path/coords/coords_{domain}_{shape}.nc
            unstack_fill_nan:
                coords: /some_path/coords/coords_{domain}_{shape}.nc

    Returns
    -------
    xr.Dataset
      Same as `ds`, but all dimensions of mask have been stacked to a single `new_dim`.
      Indexes where mask is False have been dropped.

    See Also
    --------
    unstack_fill_nan : The inverse operation.
    """
    if isinstance(mask, xr.DataArray):
        mask_1d = mask.stack({new_dim: mask.dims})
        out = ds.stack({new_dim: mask.dims}).where(mask_1d, drop=True)
    else:
        mask = ds.coords.to_dataset().drop_vars(
            [v for v in ds.coords if not any(d in mask for d in ds[v].dims)]
        )
        mask = xr.DataArray(
            np.ones(list(mask.sizes.values())), dims=mask.dims, coords=mask.coords
        )  # Make it a DataArray to fit the rest of the function
        out = ds.stack({new_dim: mask.dims}).dropna(new_dim, how="all")
    out = out.reset_index(new_dim)
    for dim in mask.dims:
        out[dim].attrs.update(ds[dim].attrs)

    original_shape = "x".join(map(str, mask.shape))

    if to_file is not None:
        # Set default path to store the information necessary to unstack
        # The name includes the domain and the original shape to uniquely identify the dataset
        domain = ds.attrs.get("cat:domain", "unknown")
        to_file = to_file.format(domain=domain, shape=original_shape)
        if not Path(to_file).parent.exists():
            Path(to_file).parent.mkdir(exist_ok=True)
        # Add all coordinates that might have been affected by the stack
        mask = mask.assign_coords(
            {c: ds[c] for c in ds.coords if any(d in mask.dims for d in ds[c].dims)}
        )
        mask.coords.to_dataset().to_netcdf(to_file)

    # Carry information about original shape to be able to unstack properly
    for dim in mask.dims:
        out[dim].attrs["original_shape"] = original_shape

        # this is needed to fix a bug in xarray '2022.6.0'
        out[dim] = xr.DataArray(
            out[dim].values,
            dims=out[dim].dims,
            coords=out[dim].coords,
            attrs=out[dim].attrs,
        )

    return out


@parse_config
def unstack_fill_nan(
    ds: xr.Dataset,
    *,
    dim: str = "loc",
    coords: None | (
        str | os.PathLike | Sequence[str | os.PathLike] | dict[str, xr.DataArray]
    ) = None,
):
    """Unstack a Dataset that was stacked by :py:func:`stack_drop_nans`.

    Parameters
    ----------
    ds : xr.Dataset
      A dataset with some dimensions stacked by `stack_drop_nans`.
    dim : str
      The dimension to unstack, same as `new_dim` in `stack_drop_nans`.
    coords : string or os.PathLike or Sequence or dict, optional
        Additional information used to reconstruct coordinates that might have been lost in the stacking (e.g., if a lat/lon grid was all NaNs).
        If a string or os.PathLike : Path to a dataset containing only those coordinates, such as the output of `to_file` in `stack_drop_nans`.
        This is the recommended option.
        If a dictionary : A mapping from the name of the coordinate that was stacked to a DataArray. Better alternative if no file is available.
        If a sequence : The names of the original dimensions that were stacked. Worst option.
        If None (default), same as a sequence, but all coordinates that have `dim` as a single dimension are used as the new dimensions.
        See Notes for more information.

    Returns
    -------
    xr.Dataset
      Same as `ds`, but `dim` has been unstacked to coordinates in `coords`.
      Missing elements are filled according to the defaults of `fill_value` of :py:meth:`xarray.Dataset.unstack`.

    Notes
    -----
    Some information might have been completely lost in the stacking process, for example, if a longitude is NaN across all latitudes.
    It is impossible to recover that information when using `coords` as a list, which is why it is recommended to use a file or a dictionary instead.

    If a dictionary is used, the keys must be the names of the coordinates that were stacked and the values must be the DataArrays.
    This method can recover both dimensions and additional coordinates that were not dimensions in the original dataset, but were stacked.

    If the original stacking was done with `stack_drop_nans` and the `to_file` argument was used, the `coords` argument should be a string with
    the path to the file. Additionally, the file name can contain the formatting fields {shape} and {domain}, which will be automatically  filled
    with the original shape of the dataset and the global attribute 'cat:domain'. If using that dynamic path, it is recommended to fill the
    argument in the xscen config.
    E.g.:

          utils:
            stack_drop_nans:
                to_file: /some_path/coords/coords_{domain}_{shape}.nc
            unstack_fill_nan:
                coords: /some_path/coords/coords_{domain}_{shape}.nc
    """
    if coords is None:
        logger.info("Dataset unstacked using no coords argument.")
        coords = [d for d in ds.coords if ds[d].dims == (dim,)]

    if isinstance(coords, str | os.PathLike):
        # find original shape in the attrs of one of the dimension
        original_shape = "unknown"
        for c in ds.coords:
            if "original_shape" in ds[c].attrs:
                original_shape = ds[c].attrs["original_shape"]
        domain = ds.attrs.get("cat:domain", "unknown")
        coords = coords.format(domain=domain, shape=original_shape)
        msg = f"Dataset unstacked using {coords}."
        logger.info(msg)
        coords = xr.open_dataset(coords)
        # separate coords that are dims or not
        coords_and_dims = {
            name: x for name, x in coords.coords.items() if name in coords.dims
        }
        coords_not_dims = {
            name: x for name, x in coords.coords.items() if name not in coords.dims
        }

        dims, crds = zip(
            *[
                (name, crd.load().values)
                for name, crd in ds.coords.items()
                if crd.dims == (dim,) and name in coords_and_dims
            ]
        )

        mindex_obj = pd.MultiIndex.from_arrays(crds, names=dims)
        mindex_coords = xr.Coordinates.from_pandas_multiindex(mindex_obj, dim)

        out = ds.drop_vars(dims).assign_coords(mindex_coords).unstack(dim)

        # only reindex with the dims
        out = out.reindex(**coords_and_dims)
        # add back the coords that aren't dims
        for c in coords_not_dims:
            out[c] = coords[c]
    else:
        coord_not_dim = {}
        # Special case where the dictionary contains both dimensions and other coordinates
        if isinstance(coords, dict):
            coord_not_dim = {
                k: v
                for k, v in coords.items()
                if len(set(v.dims).intersection(list(coords))) != 1
            }
            coords = deepcopy(coords)
            coords = {
                k: v
                for k, v in coords.items()
                if k in set(coords).difference(coord_not_dim)
            }

        dims, crds = zip(
            *[
                (name, crd.load().values)
                for name, crd in ds.coords.items()
                if (crd.dims == (dim,) and name in set(coords))
            ]
        )

        # Reconstruct the dimensions
        mindex_obj = pd.MultiIndex.from_arrays(crds, names=dims)
        mindex_coords = xr.Coordinates.from_pandas_multiindex(mindex_obj, dim)

        out = ds.drop_vars(dims).assign_coords(mindex_coords).unstack(dim)

        if isinstance(coords, dict):
            # Reindex with the coords that were dimensions
            out = out.reindex(**coords)
            # Add back the coordinates that aren't dimensions
            for c in coord_not_dim:
                out[c] = coord_not_dim[c]

        # Reorder the dimensions to match the CF conventions
        order = [out.cf.axes.get(d, [""])[0] for d in ["T", "Z", "Y", "X"]]
        order = [d for d in order if d] + [d for d in out.dims if d not in order]
        out = out.transpose(*order)

    for dim in dims:
        out[dim].attrs.update(ds[dim].attrs)

    return out


def natural_sort(_list: list[str]):
    """For strings of numbers. alternative to sorted() that detects a more natural order.

    e.g. [r3i1p1, r1i1p1, r10i1p1] is sorted as [r1i1p1, r3i1p1, r10i1p1] instead of [r10i1p1, r1i1p1, r3i1p1]
    """
    convert = lambda text: int(text) if text.isdigit() else text.lower()  # noqa: E731
    alphanum_key = lambda key: [  # noqa: E731
        convert(c) for c in re.split("([0-9]+)", key)
    ]
    return sorted(_list, key=alphanum_key)


def get_cat_attrs(
    ds: xr.Dataset | xr.DataArray | dict, prefix: str = "cat:", var_as_str=False
) -> dict:
    """Return the catalog-specific attributes from a dataset or dictionary.

    Parameters
    ----------
    ds : xr.Dataset, dict
        Dataset to be parsed. If a dictionary, it is assumed to be the attributes of the dataset (ds.attrs).
    prefix : str
        Prefix automatically generated by intake-esm. With xscen, this should be 'cat:'
    var_as_str : bool
        If True, 'variable' will be returned as a string if there is only one.

    Returns
    -------
    dict
        Compilation of all attributes in a dictionary.
    """
    if isinstance(ds, xr.Dataset | xr.DataArray):
        attrs = ds.attrs
    else:
        attrs = ds
    facets = {
        k[len(prefix) :]: v for k, v in attrs.items() if k.startswith(f"{prefix}")
    }

    # to be usable in a path
    if (
        var_as_str
        and "variable" in facets
        and not isinstance(facets["variable"], str)
        and len(facets["variable"]) == 1
    ):
        facets["variable"] = facets["variable"][0]
    return facets


def strip_cat_attrs(ds: xr.Dataset, prefix: str = "cat:"):
    """Remove attributes added from the catalog by `to_dataset` or `extract_dataset`."""
    dsc = ds.copy()
    for k in list(dsc.attrs):
        if k.startswith(prefix):
            del dsc.attrs[k]
    return dsc


@parse_config
def maybe_unstack(
    ds: xr.Dataset,
    dim: str | None = None,
    coords: str | None = None,
    rechunk: dict | None = None,
    stack_drop_nans: bool = False,
) -> xr.Dataset:
    """If stack_drop_nans is True, unstack and rechunk.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to unstack.
    dim : str, optional
        Dimension to unstack.
    coords : str, optional
        Path to a dataset containing the coords to unstack (and only those).
    rechunk : dict, optional
        If not None, rechunk the dataset after unstacking.
    stack_drop_nans : bool
        If True, unstack the dataset and rechunk it.
        If False, do nothing.

    Returns
    -------
    xr.Dataset
        Unstacked dataset.
    """
    if stack_drop_nans:
        ds = unstack_fill_nan(ds, dim=dim, coords=coords)
        if rechunk is not None:
            ds = ds.chunk(rechunk)
    return ds


# Read CVs and fill a virtual module
CV = ModuleType(
    "CV",
    (
        """
        Mappings of (controlled) vocabulary. This module is generated automatically
        from json files in xscen/CVs. Functions are essentially mappings, most of
        which are meant to provide translations between columns.\n\n
        Json files must be shallow dictionaries to be supported. If the json file
        contains a ``is_regex: True`` entry, then the keys are automatically
        translated as regex patterns and the function returns the value of the first
        key that matches the pattern. Otherwise the function essentially acts like a
        normal dictionary. The 'raw' data parsed from the json file is added in the
        ``dict`` attribute of the function.
        Example:

        .. code-block:: python

            xs.utils.CV.frequency_to_timedelta.dict

        .. literalinclude:: ../src/xscen/CVs/frequency_to_timedelta.json
           :language: json
           :caption: frequency_to_timedelta

        .. literalinclude:: ../src/xscen/CVs/frequency_to_xrfreq.json
           :language: json
           :caption: frequency_to_xrfreq

        .. literalinclude:: ../src/xscen/CVs/infer_resolution.json
           :language: json
           :caption: infer_resolution

        .. literalinclude:: ../src/xscen/CVs/resampling_methods.json
           :language: json
           :caption: resampling_methods

        .. literalinclude:: ../src/xscen/CVs/variable_names.json
           :language: json
           :caption: variable_names

        .. literalinclude:: ../src/xscen/CVs/xrfreq_to_frequency.json
           :language: json
           :caption: xrfreq_to_frequency

        .. literalinclude:: ../src/xscen/CVs/xrfreq_to_timedelta.json
           :language: json
           :caption: xrfreq_to_timedelta


        """
    ),
)


def __read_CVs(cvfile):  # noqa: N802
    with cvfile.open("r") as f:
        cv = json.load(f)
    is_regex = cv.pop("is_regex", False)
    doc = """Controlled vocabulary mapping from {name}.

    The raw dictionary can be accessed by the dict attribute of this function.

    Parameters
    ----------
    key : str
        The value to translate.{regex}
    default : 'pass', 'error' or Any
        If the key is not found in the mapping, default controls the behaviour.

        - "error", a KeyError is raised (default).
        - "pass", the key is returned.
        - another value, that value is returned.
"""

    def cvfunc(key, default="error"):
        if is_regex:
            for cin, cout in cv.items():
                try:
                    if re.fullmatch(cin, key):
                        return cout
                except TypeError:
                    pass
        else:
            if key in cv:
                return cv[key]
        if isinstance(default, str):
            if default == "pass":
                return key
            if default == "error":
                raise KeyError(key)
        return default

    cvfunc.__name__ = cvfile.stem
    cvfunc.__doc__ = doc.format(
        name=cvfile.stem.replace("_", " "),
        regex=" The key will be matched using regex" if is_regex else "",
    )
    cvfunc.__dict__["dict"] = cv
    cvfunc.__module__ = "xscen.CV"
    return cvfunc


for cvfile in Path(__file__).parent.joinpath("CVs").glob("*.json"):
    try:
        CV.__dict__[cvfile.stem] = __read_CVs(cvfile)
    # FIXME: This is a catch-all, but we should be more specific
    except Exception as err:  # noqa: BLE001
        raise ValueError(f"While reading {cvfile} got {err}")


def change_units(ds: xr.Dataset, variables_and_units: dict) -> xr.Dataset:
    """Change units of Datasets to non-CF units.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use
    variables_and_units : dict
        Description of the variables and units to output

    Returns
    -------
    xr.Dataset

    See Also
    --------
    xclim.core.units.convert_units_to, xclim.core.units.rate2amount
    """
    with xr.set_options(keep_attrs=True):
        for v in variables_and_units:
            if (v in ds) and (
                units.units2pint(ds[v]) != units.units2pint(variables_and_units[v])
            ):
                time_in_ds = units.units2pint(ds[v]).dimensionality.get("[time]")
                time_in_out = units.units2pint(
                    variables_and_units[v]
                ).dimensionality.get("[time]")

                if time_in_ds == time_in_out:
                    ds = ds.assign(
                        {v: units.convert_units_to(ds[v], variables_and_units[v])}
                    )
                elif time_in_ds - time_in_out == 1:
                    # ds is an amount
                    ds = ds.assign(
                        {v: units.amount2rate(ds[v], out_units=variables_and_units[v])}
                    )
                elif time_in_ds - time_in_out == -1:
                    # ds is a rate
                    ds = ds.assign(
                        {v: units.rate2amount(ds[v], out_units=variables_and_units[v])}
                    )
                else:
                    raise ValueError(
                        f"No known transformation between {ds[v].units} and {variables_and_units[v]} (temporal dimensionality mismatch)."
                    )
            elif (v in ds) and (ds[v].units != variables_and_units[v]):
                # update unit name if physical units are equal but not their name (ex. degC vs °C)
                ds = ds.assign({v: ds[v].assign_attrs(units=variables_and_units[v])})

    return ds


def clean_up(  # noqa: C901
    ds: xr.Dataset,
    *,
    variables_and_units: dict | None = None,
    convert_calendar_kwargs: dict | None = None,
    missing_by_var: dict | None = None,
    maybe_unstack_dict: dict | None = None,
    round_var: dict | None = None,
    common_attrs_only: None | (dict | list[xr.Dataset | str | os.PathLike]) = None,
    common_attrs_open_kwargs: dict | None = None,
    attrs_to_remove: dict | None = None,
    remove_all_attrs_except: dict | None = None,
    add_attrs: dict | None = None,
    change_attr_prefix: str | dict | None = None,
    to_level: str | None = None,
) -> xr.Dataset:
    """Clean up of the dataset.

    It can:
     - convert to the right units using xscen.utils.change_units
     - convert the calendar and interpolate over missing dates
     - call the xscen.utils.maybe_unstack function
     - round variables
     - remove a list of attributes
     - remove everything but a list of attributes
     - add attributes
     - change the prefix of the catalog attrs

    in that order.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset to clean up
    variables_and_units : dict, optional
        Dictionary of variable to convert. e.g. {'tasmax': 'degC', 'pr': 'mm d-1'}
    convert_calendar_kwargs : dict, optional
        Dictionary of arguments to feed to xarray.Dataset.convert_calendar. This will be the same for all variables.
        If missing_by_vars is given, it will override the 'missing' argument given here.
        Eg. {'calendar': 'standard', 'align_on': 'random'}
    missing_by_var : dict, optional
        Dictionary where the keys are the variables and the values are the argument to feed the `missing`
        parameters of xarray.Dataset.convert_calendar for the given variable with the
        `convert_calendar_kwargs`. When the value of an entry is 'interpolate', the missing values will be filled
        with NaNs, then linearly interpolated over time.
    maybe_unstack_dict : dict, optional
        Dictionary to pass to xscen.common.maybe_unstack function.
        The format should be: {'coords': path_to_coord_file, 'rechunk': {'time': -1 }, 'stack_drop_nans': True}.
    round_var : dict, optional
        Dictionary where the keys are the variables of the dataset and the values are the number of
        decimal places to round to.
    common_attrs_only : dict, list of datasets, or list of paths, optional
        Dictionary of datasets or list of datasets, or path to NetCDF or Zarr files.
        Keeps only the global attributes that are the same for all datasets and generates a new id.
    common_attrs_open_kwargs : dict, optional
        Dictionary of arguments for xarray.open_dataset(). Used with common_attrs_only if given paths.
    attrs_to_remove : dict, optional
        Dictionary where the keys are the variables and the values are a list of the attrs that should be removed.
        The match is done using re.fullmatch, so the strings can be regex patterns but don't need to contain '^' or '$'.
        For global attrs, use the key 'global'.
        e.g. {'global': ['unnecessary note', 'cell.*'], 'tasmax': 'old_name'}
    remove_all_attrs_except : dict, optional
        Dictionary where the keys are the variables and the values are a list of the attrs that should NOT be removed.
        The match is done using re.fullmatch, so the strings can be regex patterns but don't need to contain '^' or '$'.
        All other attributes will be deleted. For global attrs, use the key 'global'.
        e.g. {'global': ['necessary note', '^cat:'], 'tasmax': 'new_name'}
    add_attrs : dict, optional
        Dictionary where the keys are the variables and the values are a another dictionary of attributes.
        For global attrs, use the key 'global'.
        e.g. {'global': {'title': 'amazing new dataset'}, 'tasmax': {'note': 'important info about tasmax'}}
    change_attr_prefix : str or dict, optional
        If a string, replace "cat:" in the catalog global attributes by this new string.
        If a dictionary, the key is the old prefix and the value is the new prefix.
    to_level : str, optional
        The processing level to assign to the output.

    Returns
    -------
    xr.Dataset
        Cleaned up dataset

    See Also
    --------
    xclim.core.calendar.convert_calendar
    """
    ds = ds.copy()

    if variables_and_units:
        msg = f"Converting units: {variables_and_units}"
        logger.info(msg)
        ds = change_units(ds=ds, variables_and_units=variables_and_units)
    # convert calendar
    if convert_calendar_kwargs:
        # create mask of grid point that should always be nan
        ocean = ds.isnull().all("time")
        # if missing_by_var exist make sure missing data are added to time axis
        if missing_by_var:
            if not all(k in missing_by_var.keys() for k in ds.data_vars):
                raise ValueError(
                    "All variables must be in 'missing_by_var' if using this option."
                )
            convert_calendar_kwargs["missing"] = -9999

        # make default `align_on`='`random` when the initial calendar is 360day
        if (
            any(
                cal == "360_day"
                for cal in [ds.time.dt.calendar, convert_calendar_kwargs["calendar"]]
            )
            and "align_on" not in convert_calendar_kwargs
        ):
            convert_calendar_kwargs["align_on"] = "random"

        msg = f"Converting calendar with {convert_calendar_kwargs}."
        logger.info(msg)
        ds = ds.convert_calendar(**convert_calendar_kwargs).where(~ocean)

        # convert each variable individually
        if missing_by_var:
            # remove 'missing' argument to be replaced by `missing_by_var`
            del convert_calendar_kwargs["missing"]
            for var, missing in missing_by_var.items():
                msg = f"Filling missing {var} with {missing}"
                logging.info(msg)
                if missing == "interpolate":
                    ds_with_nan = ds[var].where(ds[var] != -9999)
                    converted_var = ds_with_nan.chunk({"time": -1}).interpolate_na(
                        "time", method="linear"
                    )
                else:
                    var_attrs = ds[var].attrs
                    converted_var = xr.where(ds[var] == -9999, missing, ds[var])
                    converted_var.attrs = var_attrs
                ds[var] = converted_var

    # unstack nans
    if maybe_unstack_dict:
        ds = maybe_unstack(ds, **maybe_unstack_dict)

    if round_var:
        for var, n in round_var.items():
            ds[var] = ds[var].round(n)
            new_history = (
                f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                f"Rounded '{var}' to {n} decimals."
            )
            history = (
                f"{new_history}\n{ds[var].attrs['history']}"
                if "history" in ds[var].attrs
                else new_history
            )
            ds[var].attrs["history"] = history

    if common_attrs_only:
        from .catalog import generate_id

        common_attrs_open_kwargs = common_attrs_open_kwargs or {}
        if isinstance(common_attrs_only, dict):
            common_attrs_only = list(common_attrs_only.values())

        for i in range(len(common_attrs_only)):
            if isinstance(common_attrs_only[i], str | os.PathLike):
                dataset = xr.open_dataset(
                    common_attrs_only[i], **common_attrs_open_kwargs
                )
            else:
                dataset = common_attrs_only[i]
            attributes = ds.attrs.copy()
            for a_key, a_val in attributes.items():
                if (
                    (a_key not in dataset.attrs)
                    or (a_key in ["cat:date_start", "cat:date_end"])
                    or (a_val != dataset.attrs[a_key])
                ):
                    del ds.attrs[a_key]

        # generate a new id
        try:
            ds.attrs["cat:id"] = generate_id(ds).iloc[0]
        except IndexError as err:
            msg = f"Unable to generate a new id for the dataset. Got {err}."
            logger.warning(msg)

    if to_level:
        ds.attrs["cat:processing_level"] = to_level

    # remove attrs
    if attrs_to_remove:
        for var, list_of_attrs in attrs_to_remove.items():
            obj = ds if var == "global" else ds[var]
            to_remove = list(
                chain.from_iterable(
                    [
                        list(filter(re.compile(attr).fullmatch, list(obj.attrs.keys())))
                        for attr in list_of_attrs
                    ]
                )
            )
            for attr in to_remove:
                del obj.attrs[attr]

    # delete all attrs, but the ones in the list
    if remove_all_attrs_except:
        for var, list_of_attrs in remove_all_attrs_except.items():
            obj = ds if var == "global" else ds[var]
            to_keep = list(
                chain.from_iterable(
                    [
                        list(filter(re.compile(attr).fullmatch, list(obj.attrs.keys())))
                        for attr in list_of_attrs
                    ]
                )
            )
            to_remove = list(set(obj.attrs.keys()).difference(to_keep))
            for attr in to_remove:
                del obj.attrs[attr]

    if add_attrs:
        for var, attrs in add_attrs.items():
            obj = ds if var == "global" else ds[var]
            for attrname, attrtmpl in attrs.items():
                obj.attrs[attrname] = attrtmpl

    if change_attr_prefix:
        if isinstance(change_attr_prefix, str):
            change_attr_prefix = {"cat:": change_attr_prefix}
        # Make sure that the prefixes are in the right format
        chg_attr_prefix = {}
        for old_prefix, new_prefix in change_attr_prefix.items():
            if not old_prefix.endswith(":"):
                old_prefix += ":"
            if not new_prefix.endswith(":"):
                new_prefix += ":"
            chg_attr_prefix[old_prefix] = new_prefix

        # Change the prefixes, but keep the order of the keys
        attrs = {}
        for ds_attr in list(ds.attrs.keys()):
            changed = False
            for old_prefix, new_prefix in chg_attr_prefix.items():
                if ds_attr.startswith(old_prefix):
                    new_name = ds_attr.replace(old_prefix, new_prefix)
                    attrs[new_name] = ds.attrs[ds_attr]
                    changed = True
            if not changed:
                attrs[ds_attr] = ds.attrs[ds_attr]
        ds.attrs = attrs

    return ds


def unstack_dates(  # noqa: C901
    ds: xr.Dataset,
    seasons: dict[int, str] | None = None,
    new_dim: str | None = None,
    winter_starts_year: bool = False,
):
    """Unstack a multi-season timeseries into a yearly axis and a season one.

    Parameters
    ----------
    ds : xr.Dataset or DataArray
      The xarray object with a "time" coordinate.
      Only supports monthly or coarser frequencies.
      The time axis must be complete and regular (`xr.infer_freq(ds.time)` doesn't fail).
    seasons : dict, optional
      A dictionary from month number (as int) to a season name.
      If not given, it is guessed from the time coordinate frequency.
      See notes.
    new_dim : str, optional
      The name of the new dimension.
      If None, the name is inferred from the frequency of the time axis.
      See notes.
    winter_starts_year : bool
      If True, the year of winter (DJF) is built from the year of January, not December.
      i.e. DJF made from [Dec 1980, Jan 1981, and Feb 1981] will be associated with the year 1981, not 1980.

    Returns
    -------
    xr.Dataset or DataArray
      Same as ds but the time axis is now yearly (YS-JAN) and the seasons are along the new dimension.

    Notes
    -----
    When `seasons` is None, the inferred frequency determines the new coordinate:
    - For MS, the coordinates are the month abbreviations in english (JAN, FEB, etc.)
    - For ?QS-? and other ?MS frequencies, the coordinates are the initials of the months in each season. Ex: QS -> DJF, MAM, JJA, SON.
    - For YS or YS-JAN, the new coordinate has a single value of "annual".
    - For ?YS-? frequencies, the new coordinate has a single value of "annual-{anchor}". Ex: YS-JUL -> "annual-JUL".

    When `new_dim` is None, the new dimension name is inferred from the frequency:
    - For ?YS, ?QS frequencies or ?MS with mult > 1, the new dimension is "season".
    - For MS, the new dimension is "month".

    """
    # Get some info about the time axis
    freq = xr.infer_freq(ds.time)
    if freq is None:
        raise ValueError(
            "The data must have a clean time coordinate. If you know the "
            "data's frequency, please pass `ds.resample(time=freq).first()` "
            "to pad missing dates and reset the time coordinate."
        )
    first, last = ds.indexes["time"][[0, -1]]
    use_cftime = xr.coding.times.contains_cftime_datetimes(ds.time.variable)
    calendar = ds.time.dt.calendar
    mult, base, isstart, anchor = parse_offset(freq)

    if base not in "YAQM":
        raise ValueError(
            f"Only monthly frequencies or coarser are supported. Got: {freq}."
        )

    if new_dim is None:
        if base == "M" and mult == 1:
            new_dim = "month"
        else:
            new_dim = "season"

    if base in "YA":
        if seasons:
            seaname = f"{seasons[first.month]}"
        elif anchor == "JAN":
            seaname = "annual"
        else:
            seaname = f"annual-{anchor}"
        if mult > 1:
            seaname = f"{mult}{seaname}"
        # Fast track for annual, if nothing more needs to be done.
        if winter_starts_year is False:
            dso = ds.expand_dims({new_dim: [seaname]})
            dso["time"] = xr.date_range(
                f"{first.year}-01-01",
                f"{last.year}-01-01",
                freq=f"{mult}YS",
                calendar=calendar,
                use_cftime=use_cftime,
            )
            return dso
        else:
            seasons = seasons or {}
            seasons.update({first.month: seaname})

    if base == "M" and 12 % mult != 0:
        raise ValueError(
            f"Only periods that divide the year evenly are supported. Got {freq}."
        )

    # Guess the new season coordinate
    if seasons is None:
        if base == "Q" or (base == "M" and mult > 1):
            # Labels are the month initials
            months = np.array(list("JFMAMJJASOND"))
            n = mult * {"M": 1, "Q": 3}[base]
            seasons = {
                m: "".join(months[np.array(range(m - 1, m + n - 1)) % 12])
                for m in np.unique(ds.time.dt.month)
            }
        else:  # M or MS
            seasons = xr.coding.cftime_offsets._MONTH_ABBREVIATIONS
    else:
        # Only keep the entries for the months in the data
        seasons = {m: seasons[m] for m in np.unique(ds.time.dt.month)}
    # The ordered season names
    seas_list = [seasons[month] for month in sorted(seasons.keys())]

    # Multi-month seasons that isn't synced with january
    if winter_starts_year and 1 not in seasons:
        # The last label is the period that overlaps the year
        winter_month = max(seasons.keys())
        # Put it back in the beginning
        seas_list = [seasons[winter_month]] + seas_list[:-1]
        # The year associated with each timestamp (add 1 in winter)
        years = ds.time.dt.year + xr.where(ds.time.dt.month == winter_month, 1, 0)
    else:  # Monthly or aligned seasons
        years = ds.time.dt.year

    # The goal here is to use `reshape()` instead of `unstack` to limit the number of dask operations.
    # Thus, the time axis must be properly constructed so that reshapes fits the final size.
    # We pad on both sides to ensure full years
    pad_left = seas_list.index(seasons[first.month])
    pad_right = len(seas_list) - (seas_list.index(seasons[last.month]) + 1)
    dsp = ds.pad(time=(pad_left, pad_right))  # pad with NaN
    # Similarly pad our "group labels".
    years = years.pad(time=(pad_left, pad_right), constant_values=(years[0], years[-1]))

    # New coords
    new_time = xr.date_range(  # New time axis (YS)
        f"{years[0].item()}-01-01",
        f"{years[-1].item()}-01-01",
        freq="YS",
        calendar=calendar,
        use_cftime=use_cftime,
    )

    def reshape_da(da):
        # Replace (A,'time',B) by (A,'time', 'season',B) in both the new shape and the new dims
        new_dims = list(
            chain.from_iterable(
                [d] if d != "time" else ["time", new_dim] for d in da.dims
            )
        )
        new_shape = [len(new_coords[d]) for d in new_dims]
        # Use dask or numpy's algo.

        if uses_dask(da):
            # This is where it happens. Flox will minimally rechunk
            # so the reshape operation can be performed blockwise
            da = flox.xarray.rechunk_for_blockwise(da, "time", years)
        return xr.DataArray(da.data.reshape(new_shape), dims=new_dims)

    new_coords = dict(ds.coords)
    new_coords.update({"time": new_time, new_dim: seas_list})

    # put other coordinates that depend on time in the new shape
    for coord in new_coords:
        if (coord not in ["time", new_dim]) and ("time" in ds[coord].dims):
            new_coords[coord] = reshape_da(dsp[coord])

    if isinstance(ds, xr.Dataset):
        dso = dsp.map(reshape_da, keep_attrs=True)
    else:
        dso = reshape_da(dsp)
    return dso.assign_coords(**new_coords)


def ensure_correct_time(ds: xr.Dataset, xrfreq: str) -> xr.Dataset:
    """Ensure a dataset has the correct time coordinate, as expected for the given frequency.

    Daily or finer datasets are "floored" even if `xr.infer_freq` succeeds.
    Errors are raised if the number of data points per period is not 1.
    The dataset is modified in-place, but returned nonetheless.
    """
    # Check if we got the expected freq (skip for too short timeseries)
    inffreq = xr.infer_freq(ds.time) if ds.time.size > 2 else None
    if inffreq == xrfreq:
        # Even when the freq is correct, we ensure the correct "anchor" for daily and finer
        if xrfreq in "DHTMUL":
            ds["time"] = ds.time.dt.floor(xrfreq)
    else:
        # We can't infer it, there might be a problem
        counts = ds.time.resample(time=xrfreq).count()
        if (counts > 1).any().item():
            raise ValueError(
                "Dataset is labelled as having a sampling frequency of "
                f"{xrfreq}, but some periods have more than one data point."
            )
        if (counts.isnull() | (counts == 0)).any().item():
            raise ValueError(
                "The resampling count contains NaNs or 0s. There might be some missing data."
            )
        ds["time"] = counts.time
    return ds


def standardize_periods(
    periods: list[str | pd.Timestamp] | list[list[str | pd.Timestamp]] | None,
    multiple: bool = True,
    end_of_periods: bool = True,
    out_dtype: str = "str",
) -> list[str] | list[list[str]] | None:
    """Reformats the input to a list of strings or Timestamps, ['start', 'end'], or a list of such lists. Does not modify in-place.

    Parameters
    ----------
    periods : list of str or pd.Timestamp, or list of lists of str or pd.Timestamp, optional
      The period(s) to standardize. If None, return None.
    multiple : bool
        If True, return a list of periods, otherwise return a single period.
    end_of_periods: bool or str
        If 'YE' or 'ME', the returned date will be the end of the year or month that contains the received date.
        If True (default), standardizes yearly and monthly periods to end on the last second of the last day of the year/month.
        This parameter is only used for str periods that do not specify the month/day.
    out_dtype : str
        Choices are 'datetime', 'period' or 'str'. Defaults to 'str', which will only output the year.
    """
    if periods is None:
        return periods

    periods = deepcopy(periods)
    if not isinstance(periods[0], list):
        periods = [periods]

    for i in range(len(periods)):
        if len(periods[i]) != 2:
            raise ValueError(
                "Each instance of 'periods' should be comprised of two elements: [start, end]."
            )
        period = periods[i]
        if isinstance(period[0], int) or isinstance(period[0], str):
            period[0] = date_parser(str(period[0]), out_dtype="datetime")
        if isinstance(period[1], int) or isinstance(period[1], str):
            period[1] = date_parser(
                str(period[1]), out_dtype="datetime", end_of_period=end_of_periods
            )
        if period[0] > period[1]:
            raise ValueError(
                f"'periods' should be in chronological order, received {periods[i]}."
            )
        # TODO: allow more than year in periods for out_dtype = str
        periods[i] = [
            date_parser(period[0], out_dtype=out_dtype, strtime_format="%Y"),
            date_parser(period[1], out_dtype=out_dtype, strtime_format="%Y"),
        ]
    if multiple:
        return periods
    else:
        if len(periods) > 1:
            raise ValueError(
                f"'period' should be a single instance of [start, end], received {len(periods)}."
            )
        return periods[0]


def season_sort_key(idx: pd.Index, name: str | None = None):
    """Get a proper sort key for a "season" or "month" index to avoid alphabetical sorting.

    If any of the values in the index is not recognized as a 3-letter
    season code or a 3-letter month abbreviation, the operation is
    aborted and the index is returned untouched.
    DJF is the first season of the year.

    Parameters
    ----------
    idx : pd.Index
      Any array that implements a `map` method.
      If name is "month", index elements are expected to be 3-letter month abbreviations, uppercase (JAN, FEB, etc).
      If name is "season", index elements are expected to be 3-letter season abbreviations, uppercase (DJF, AMJ, OND, etc.)
      If anything else, the index is returned untouched.
    name : str, optional
      The index name. By default, the `name` attribute of the index is used, if present.

    Returns
    -------
    idx : Integer sort key for months and seasons, the input index untouched otherwise.
    """
    try:
        if (name or getattr(idx, "name", None)) == "season":
            m = "DJFMAMJJASONDJ"
            return idx.map(m.index)
        if (name or getattr(idx, "name", None)) == "month":
            m = list(xr.coding.cftime_offsets._MONTH_ABBREVIATIONS.values())
            return idx.map(m.index)
    except (TypeError, ValueError) as err:
        # ValueError if string not in seasons, or value not in months
        # TypeError if season element was not a string.
        logging.error(err)
    return idx


def xrfreq_to_timedelta(freq: str):
    """Approximate the length of a period based on its frequency offset."""
    N, B, _, _ = parse_offset(freq)
    return N * pd.Timedelta(CV.xrfreq_to_timedelta(B, "NaT"))


def ensure_new_xrfreq(freq: str) -> str:  # noqa: C901
    """Convert the frequency string to the newer syntax (pandas >= 2.2) if needed."""
    # Copied from xarray xr.coding.cftime_offsets._legacy_to_new_freq
    # https://github.com/pydata/xarray/pull/8627/files
    if not isinstance(freq, str):
        # For when freq is NaN or None in a catalog
        return freq
    try:
        freq_as_offset = cfoff.to_offset(freq, warn=False)
    except ValueError:
        # freq may be valid in pandas but not in xarray
        return freq

    if isinstance(freq_as_offset, cfoff.MonthEnd) and "ME" not in freq:
        freq = freq.replace("M", "ME")
    elif isinstance(freq_as_offset, cfoff.QuarterEnd) and "QE" not in freq:
        freq = freq.replace("Q", "QE")
    elif isinstance(freq_as_offset, cfoff.YearBegin) and "YS" not in freq:
        freq = freq.replace("AS", "YS")
    elif isinstance(freq_as_offset, cfoff.YearEnd):
        if "A-" in freq:
            # Check for and replace "A-" instead of just "A" to prevent
            # corrupting anchored offsets that contain "Y" in the month
            # abbreviation, e.g. "A-MAY" -> "YE-MAY".
            freq = freq.replace("A-", "YE-")
        elif "Y-" in freq:
            freq = freq.replace("Y-", "YE-")
        elif freq.endswith("A"):
            # the "A-MAY" case is already handled above
            freq = freq.replace("A", "YE")
        elif "YE" not in freq and freq.endswith("Y"):
            # the "Y-MAY" case is already handled above
            freq = freq.replace("Y", "YE")
    elif isinstance(freq_as_offset, cfoff.Hour):
        freq = freq.replace("H", "h")
    elif isinstance(freq_as_offset, cfoff.Minute):
        freq = freq.replace("T", "min")
    elif isinstance(freq_as_offset, cfoff.Second):
        freq = freq.replace("S", "s")
    elif isinstance(freq_as_offset, cfoff.Millisecond):
        freq = freq.replace("L", "ms")
    elif isinstance(freq_as_offset, cfoff.Microsecond):
        freq = freq.replace("U", "us")

    return freq


def _xarray_defaults(**kwargs):
    """Translate from xscen's extract names to intake-esm names and put better defaults."""
    if "xr_open_kwargs" in kwargs:
        kwargs["xarray_open_kwargs"] = kwargs.pop("xr_open_kwargs")
    if "xr_combine_kwargs" in kwargs:
        kwargs["xarray_combine_by_coords_kwargs"] = kwargs.pop("xr_combine_kwargs")

    kwargs.setdefault("xarray_open_kwargs", {}).setdefault("chunks", {})
    kwargs.setdefault("xarray_combine_by_coords_kwargs", {}).setdefault(
        "data_vars", "minimal"
    )
    return kwargs


def rechunk_for_resample(obj: xr.DataArray | xr.Dataset, **resample_kwargs):
    if not uses_dask(obj):
        return obj

    res = obj.resample(**resample_kwargs)
    return flox.xarray.rechunk_for_blockwise(obj, res._dim, res._codes)


def publish_release_notes(*args, **kwargs):
    """Backward compatibility for the old function."""
    warnings.warn(
        "'xscen.utils.publish_release_notes' has been moved to 'xscen.testing.publish_release_notes'."
        "Support for this function will be removed in xscen v0.12.0.",
        FutureWarning,
    )

    from .testing import publish_release_notes as prn

    return prn(*args, **kwargs)


def show_versions(*args, **kwargs):
    """Backward compatibility for the old function."""
    warnings.warn(
        "'xscen.utils.show_versions' has been moved to 'xscen.testing.show_versions'."
        "Support for this function will be removed in xscen v0.12.0.",
        FutureWarning,
    )

    from .testing import show_versions as sv

    return sv(*args, **kwargs)
